import asyncio
import unittest
from unittest import mock

from aioes.connection import Connection
# 400 404 409
from aioes.exception import (TransportError, RequestError,
                             NotFoundError, ConflictError)
from aioes.transport import Endpoint
from asyncio.test_utils import TestLoop


class TestConnection(unittest.TestCase):

    def setUp(self):
        self.loop = TestLoop()

    def test_ctor(self):
        c = Connection(Endpoint('http', 'host', 9999), loop=self.loop)
        self.assertEqual(Endpoint('http', 'host', 9999), c.endpoint)

    def test_bad_status_common(self):

        @asyncio.coroutine
        def go():
            conn = Connection(Endpoint('http', 'host', 9999), loop=self.loop)
            resp = mock.Mock()
            resp.status = 401
            r2 = asyncio.Future(loop=self.loop)
            r2.set_result('{"a": 1}')
            resp.text.return_value = r2
            fut = asyncio.Future(loop=self.loop)
            fut.set_result(resp)
            conn._session.request = mock.Mock(return_value=fut)

            with self.assertRaises(TransportError) as ctx:
                yield from conn.perform_request('GET', '/data', None, None)
            self.assertEqual(401, ctx.exception.status_code)
            self.assertEqual('{"a": 1}', ctx.exception.error)
            self.assertEqual({"a": 1}, ctx.exception.info)

        self.loop.run_until_complete(go())

    def test_bad_json(self):

        @asyncio.coroutine
        def go():
            conn = Connection(Endpoint('http', 'host', 9999), loop=self.loop)
            resp = mock.Mock()
            resp.status = 401
            r2 = asyncio.Future(loop=self.loop)
            r2.set_result('error text')
            resp.text.return_value = r2
            fut = asyncio.Future(loop=self.loop)
            fut.set_result(resp)
            conn._session.request = mock.Mock(return_value=fut)

            with self.assertRaises(TransportError) as ctx:
                yield from conn.perform_request('GET', '/data', None, None)
            self.assertEqual(401, ctx.exception.status_code)
            self.assertEqual('error text', ctx.exception.error)
            self.assertEqual(None, ctx.exception.info)

        self.loop.run_until_complete(go())

    def test_bad_status_400(self):

        @asyncio.coroutine
        def go():
            conn = Connection(Endpoint('http', 'host', 9999), loop=self.loop)
            resp = mock.Mock()
            resp.status = 400
            r2 = asyncio.Future(loop=self.loop)
            r2.set_result('{"a": 1}')
            resp.text.return_value = r2
            fut = asyncio.Future(loop=self.loop)
            fut.set_result(resp)
            conn._session.request = mock.Mock(return_value=fut)

            with self.assertRaises(RequestError) as ctx:
                yield from conn.perform_request('GET', '/data', None, None)
            self.assertEqual(400, ctx.exception.status_code)
            self.assertEqual('{"a": 1}', ctx.exception.error)
            self.assertEqual({"a": 1}, ctx.exception.info)

        self.loop.run_until_complete(go())

    def test_bad_status_404(self):

        @asyncio.coroutine
        def go():
            conn = Connection(Endpoint('http', 'host', 9999), loop=self.loop)
            resp = mock.Mock()
            resp.status = 404
            r2 = asyncio.Future(loop=self.loop)
            r2.set_result('{"a": 1}')
            resp.text.return_value = r2
            fut = asyncio.Future(loop=self.loop)
            fut.set_result(resp)
            conn._session.request = mock.Mock(return_value=fut)

            with self.assertRaises(NotFoundError) as ctx:
                yield from conn.perform_request('GET', '/data', None, None)
            self.assertEqual(404, ctx.exception.status_code)
            self.assertEqual('{"a": 1}', ctx.exception.error)
            self.assertEqual({"a": 1}, ctx.exception.info)

        self.loop.run_until_complete(go())

    def test_bad_status_409(self):

        @asyncio.coroutine
        def go():
            conn = Connection(Endpoint('http', 'host', 9999), loop=self.loop)
            resp = mock.Mock()
            resp.status = 409
            r2 = asyncio.Future(loop=self.loop)
            r2.set_result('{"a": 1}')
            resp.text.return_value = r2
            fut = asyncio.Future(loop=self.loop)
            fut.set_result(resp)
            conn._session.request = mock.Mock(return_value=fut)

            with self.assertRaises(ConflictError) as ctx:
                yield from conn.perform_request('GET', '/data', None, None)
            self.assertEqual(409, ctx.exception.status_code)
            self.assertEqual('{"a": 1}', ctx.exception.error)
            self.assertEqual({"a": 1}, ctx.exception.info)

        self.loop.run_until_complete(go())



import asyncio

from .utils import NamespacedClient
from .utils import _make_path

default = object()


def _decode_text(s):
    return s


class NodesClient(NamespacedClient):

    @asyncio.coroutine
    def info(self, node_id=None, metric=None, *,
             flat_settings=default, human=default):
        """
        The cluster nodes info API allows to retrieve one or more (or all) of
        the cluster nodes information.
        """
        params = {}
        if flat_settings is not default:
            params['flat_settings'] = bool(flat_settings)
        if human is not default:
            params['human'] = bool(human)
        _, data = yield from self.transport.perform_request(
            'GET',
            _make_path('_nodes', node_id, metric),
            params=params)
        return data

    @asyncio.coroutine
    def shutdown(self, node_id=None, *, delay=default, exit=default):
        """
        The nodes shutdown API allows to shutdown one or more (or all) nodes in
        the cluster.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cluster-nodes-shutdown.html>`_

        :arg node_id: A comma-separated list of node IDs or names to perform
            the operation on; use `_local` to perform the operation on
            the node you're connected to, leave empty to perform the operation
            on all nodes
        :arg delay: Set the delay for the operation (default: 1s)
        :arg exit: Exit the JVM as well (default: true)
        """
        params = {}

        if delay is not default:
            params['delay'] = delay
        if exit is not default:
            params['exit'] = bool(exit)

        _, data = yield from self.transport.perform_request(
            'POST', _make_path('_cluster', 'nodes', node_id, '_shutdown'),
            params=params
        )
        return data

    @asyncio.coroutine
    def stats(self, node_id=None, metric=None, index_metric=None, *,
              completion_fields=default, fielddata_fields=default,
              fields=default, groups=default, human=default, level=default,
              types=default):
        """
        The cluster nodes stats API allows to retrieve one or more (or all) of
        the cluster nodes statistics.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cluster-nodes-stats.html>`_

        :arg node_id: A comma-separated list of node IDs or names to limit the
            returned information; use `_local` to return information from the
            node you're connecting to, leave empty to get information from all
            nodes
        :arg metric: Limit the information returned to the specified metrics.
            Possible options are: "_all", "breaker", "fs", "http", "indices",
            "jvm", "network", "os", "process", "thread_pool", "transport"
        :arg index_metric: Limit the information returned for `indices` metric
            to the specific index metrics. Isn't used if `indices` (or `all`)
            metric isn't specified. Possible options are: "_all", "completion",
            "docs", "fielddata", "filter_cache", "flush", "get", "id_cache",
            "indexing", "merge", "percolate", "refresh", "search", "segments",
            "store", "warmer"
        :arg completion_fields: A comma-separated list of fields
            for `fielddata` and `suggest` index metric (supports wildcards)
        :arg fielddata_fields: A comma-separated list of fields for `fielddata`
            index metric (supports wildcards)
        :arg fields: A comma-separated list of fields for `fielddata` and
            `completion` index metric (supports wildcards)
        :arg groups: A comma-separated list of search groups for `search` index
            metric
        :arg human: Whether to return time and byte values in human-readable
            format., default False
        :arg level: Return indices stats aggregated at node, index or shard
            level, default 'node'
        :arg types: A comma-separated list of document types for the `indexing`
            index metric
        """
        params = {}

        if completion_fields is not default:
            params['completion_fields'] = completion_fields
        if fielddata_fields is not default:
            params['fielddata_fields'] = fielddata_fields
        if fields is not default:
            params['fields'] = fields
        if groups is not default:
            params['groups'] = groups
        if human is not default:
            params['human'] = bool(human)
        if level is not default:
            params['level'] = level
        if types is not default:
            params['types'] = types

        _, data = yield from self.transport.perform_request(
            'GET',
            _make_path('_nodes', node_id, 'stats', metric, index_metric),
            params=params
        )
        return data

    @asyncio.coroutine
    def hot_threads(self, node_id=None, *, type_=default, interval=default,
                    snapshots=default, threads=default):
        """
        An API allowing to get the current hot threads on each node
        in the cluster.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cluster-nodes-hot-threads.html>`_

        :arg node_id: A comma-separated list of node IDs or names to limit the
            returned information; use `_local` to return information from the
            node you're connecting to, leave empty to get information from all
            nodes
        :arg type_: The type to sample (default: cpu)
        :arg interval: The interval for the second sampling of threads
        :arg snapshots: Number of samples of thread stacktrace (default: 10)
        :arg threads: Specify the number of threads to provide information for
            (default: 3)
        """
        params = {}

        if type_ is not default:
            # avoid python reserved words
            params['type'] = type_
        if interval is not default:
            params['interval'] = interval
        if snapshots is not default:
            params['snapshots'] = snapshots
        if threads is not default:
            params['threads'] = threads

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_nodes', node_id, 'hot_threads'),
            params=params, decoder=_decode_text
        )
        return data



import asyncio
import unittest
from aioes import Elasticsearch
from aioes.exception import NotFoundError


class TestNodes(unittest.TestCase):

    def setUp(self):
        self._index = 'elastic_search'
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.cl = Elasticsearch([{'host': 'localhost'}], loop=self.loop)
        self.addCleanup(self.cl.close)
        try:
            self.loop.run_until_complete(
                self.cl.delete(self._index, refresh=True))
        except NotFoundError:
            pass

    def tearDown(self):
        self.loop.close()

    def test_info(self):
        @asyncio.coroutine
        def go():
            ret = yield from self.cl.nodes.info()
            self.assertIn('cluster_name', ret)

        self.loop.run_until_complete(go())

    def test_stats(self):
        @asyncio.coroutine
        def go():
            ret = yield from self.cl.nodes.stats()
            self.assertIn('nodes', ret)
            self.assertTrue(len(ret['nodes']) > 0)

        self.loop.run_until_complete(go())

    def test_hot_threads(self):
        @asyncio.coroutine
        def go():
            ret = yield from self.cl.nodes.hot_threads()
            self.assertIn('cpu usage by thread', ret)

        self.loop.run_until_complete(go())



import asyncio

from .utils import NamespacedClient
from .utils import _make_path

default = object()


class ClusterClient(NamespacedClient):

    @asyncio.coroutine
    def health(self, index=None, *,
               level=default, local=default, master_timeout=default,
               timeout=default, wait_for_active_shards=default,
               wait_for_nodes=default, wait_for_relocating_shards=default,
               wait_for_status=default):
        """
        Get a very simple status on the health of the cluster.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/
        current/cluster-health.html>`_

        :arg index: Limit the information returned to a specific index
        :arg level: Specify the level of detail for returned information,
             default u'cluster'
        :arg local: Return local information, do not retrieve the state from
             master node (default: false)
        :arg master_timeout: Explicit operation timeout for connection to
             master node
        :arg timeout: Explicit operation timeout
        :arg wait_for_active_shards: Wait until the specified number of shards
             is active
        :arg wait_for_nodes: Wait until the specified number of nodes is
             available
        :arg wait_for_relocating_shards: Wait until the specified number of
             relocating shards is finished
        :arg wait_for_status: Wait until cluster is in a specific state,
             default None
        """
        params = {}
        if level is not default:
            if not isinstance(level, str):
                raise TypeError("'level' parameter is not a string")
            elif level.lower() in ('cluster', 'indices', 'shards'):
                params['level'] = level.lower()
            else:
                raise ValueError("'level' parameter should be one"
                                 " of 'cluster', 'indices', 'shards'")
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if timeout is not default:
            params['timeout'] = timeout
        if wait_for_active_shards is not default:
            params['wait_for_active_shards'] = int(wait_for_active_shards)
        if wait_for_nodes is not default:
            params['wait_for_nodes'] = str(wait_for_nodes)
        if wait_for_relocating_shards is not default:
            params['wait_for_relocating_shards'] = \
                int(wait_for_relocating_shards)
        if wait_for_status is not default:
            if not isinstance(wait_for_status, str):
                raise TypeError("'wait_for_status' parameter is not a string")
            elif wait_for_status.lower() in ('green', 'yellow', 'red'):
                params['wait_for_status'] = wait_for_status.lower()
            else:
                raise ValueError("'wait_for_status' parameter should be one"
                                 " of 'green', 'yellow', 'red'")

        _, data = yield from self.transport.perform_request(
            'GET',
            _make_path('_cluster', 'health', index),
            params=params)
        return data

    @asyncio.coroutine
    def pending_tasks(self, *, local=default, master_timeout=default):
        """
        The pending cluster tasks API returns a list of any cluster-level
        changes (e.g. create index, update mapping, allocate or fail shard)
        which have not yet been executed.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cluster-pending.html>`_

        :arg local: Return local information, do not retrieve the state
            from master node (default: false)
        :arg master_timeout: Specify timeout for connection to master
        """
        params = {}

        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout

        _, data = yield from self.transport.perform_request(
            'GET', '/_cluster/pending_tasks',
            params=params
        )
        return data

    @asyncio.coroutine
    def state(self, metric=None, index=None, *, index_templates=default,
              local=default, master_timeout=default, flat_settings=default):
        """
        Get a comprehensive state information of the whole cluster.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cluster-state.html>`_

        :arg metric: Limit the information returned to the specified metrics.
            Possible values: "_all", "blocks", "index_templates", "metadata",
            "nodes", "routing_table", "master_node", "version"
        :arg index: A comma-separated list of index names; use `_all` or empty
            string to perform the operation on all indices
        :arg index_templates: A comma separated list to return specific index
            templates when returning metadata.
        :arg local: Return local information, do not retrieve the state
            from master node (default: false)
        :arg master_timeout: Specify timeout for connection to master
        :arg flat_settings: Return settings in flat format (default: false)
        """
        params = {}

        if local is not default:
            params['local'] = bool(local)
        if index_templates is not default:
            params['index_templates'] = index_templates
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if flat_settings is not default:
            params['flat_settings'] = bool(flat_settings)

        if index and not metric:
            metric = '_all'

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cluster', 'state', metric, index),
            params=params
        )
        return data

    @asyncio.coroutine
    def stats(self, node_id=None, *, flat_settings=default, human=default):
        """
        The Cluster Stats API allows to retrieve statistics from a cluster wide
        perspective. The API returns basic index metrics and information about
        the current nodes that form the cluster.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cluster-stats.html>`_

        :arg node_id: A comma-separated list of node IDs or names to limit the
            returned information; use `_local` to return information from the
            node you're connecting to, leave empty to get information from
            all nodes
        :arg flat_settings: Return settings in flat format (default: false)
        :arg human: Whether to return time and byte values in
            human-readable format.

        """
        params = {}

        if flat_settings is not default:
            params['flat_settings'] = bool(flat_settings)
        if human is not default:
            params['human'] = bool(human)

        url = '/_cluster/stats'
        if node_id:
            url = _make_path('_cluster/stats/nodes', node_id)
        _, data = yield from self.transport.perform_request(
            'GET', url, params=params
        )
        return data

    @asyncio.coroutine
    def reroute(self, body=None, *, dry_run=default, explain=default,
                filter_metadata=default, master_timeout=default,
                timeout=default):
        """
        Explicitly execute a cluster reroute allocation command including
        specific commands.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cluster-reroute.html>`_

        :arg body: The definition of `commands` to perform
            (`move`, `cancel`, `allocate`)
        :arg dry_run: Simulate the operation only and return
            the resulting state
        :arg explain: Return an explanation of why the commands can or
            cannot be executed
        :arg filter_metadata: Don't return cluster state metadata
            (default: false)
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        :arg timeout: Explicit operation timeout
        """
        params = {}

        if dry_run is not default:
            params['dry_run'] = bool(dry_run)
        if explain is not default:
            params['explain'] = bool(explain)
        if filter_metadata is not default:
            params['filter_metadata'] = bool(filter_metadata)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if timeout is not default:
            params['timeout'] = timeout

        _, data = yield from self.transport.perform_request(
            'POST', '/_cluster/reroute', params=params, body=body
        )
        return data

    @asyncio.coroutine
    def get_settings(self, *, flat_settings=default, master_timeout=default,
                     timeout=default):
        """
        Get cluster settings.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cluster-update-settings.html>`_

        :arg flat_settings: Return settings in flat format (default: false)
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        :arg timeout: Explicit operation timeout
        """
        params = {}

        if flat_settings is not default:
            params['flat_settings'] = bool(flat_settings)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if timeout is not default:
            params['timeout'] = timeout

        _, data = yield from self.transport.perform_request(
            'GET', '/_cluster/settings', params=params
        )
        return data

    @asyncio.coroutine
    def put_settings(self, body, *, flat_settings=default):
        """
        Update cluster wide specific settings.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cluster-update-settings.html>`_

        :arg body: The settings to be updated. Can be either `transient` or
            `persistent` (survives cluster restart).
        :arg flat_settings: Return settings in flat format (default: false)
        """
        params = {}

        if flat_settings is not default:
            params['flat_settings'] = bool(flat_settings)

        _, data = yield from self.transport.perform_request(
            'PUT', '/_cluster/settings',
            params=params, body=body
        )
        return data



from datetime import date, datetime
from urllib.parse import quote_plus

# parts of URL to be omitted
SKIP_IN_PATH = (None, b'', [], ())


def _escape(value):
    """
    Escape a single value of a URL string or a query parameter. If it is a list
    or tuple, turn it into a comma-separated string first.
    """

    # make sequences into comma-separated stings
    if isinstance(value, (list, tuple)):
        value = ','.join(value)

    # dates and datetimes into isoformat
    elif isinstance(value, (date, datetime)):
        value = value.isoformat()

    # make bools into true/false strings
    elif isinstance(value, bool):
        value = str(value).lower()

    # encode strings to utf-8
    if isinstance(value, (str, bytes)):
        try:
            return value.encode('utf-8')
        except UnicodeDecodeError:
            # Python 2 and str, no need to re-encode
            pass

    return str(value)


def _make_path(*parts):
    """
    Create a URL string from parts, omit all `None` values and empty strings.
    Convert lists and tuples to comma separated values.
    """
    # TODO: maybe only allow some parts to be lists/tuples ?
    # preserve ',' and '*' in url for nicer URLs in logs
    return '/' + '/'.join(
        quote_plus(_escape(p), b',*')
        for p in parts if p not in SKIP_IN_PATH)


class NamespacedClient:

    def __init__(self, client):
        self._client = client

    @property
    def client(self):
        return self._client

    @property
    def transport(self):
        return self._client.transport



import abc
import asyncio
import collections
import random
import time

from .log import logger


class AbstractSelector(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def select(self, connections):
        pass  # pragma: no cover


class RandomSelector(AbstractSelector):

    def select(self, connections):
        return random.choice(connections)


class RoundRobinSelector(AbstractSelector):
    def __init__(self):
        self._current = 0

    def select(self, connections):
        self._current += 1
        if self._current >= len(connections):
            self._current = 0
        return connections[self._current]


class ConnectionPool:
    def __init__(self, connections, *, dead_timeout=60, timeout_cutoff=5,
                 selector_factory=RoundRobinSelector,
                 loop):
        self._dead_timeout = dead_timeout
        self._timeout_cutoff = timeout_cutoff
        self._selector = selector_factory()
        self._dead = asyncio.PriorityQueue(len(connections), loop=loop)
        self._dead_count = collections.Counter()
        self._connections = connections
        self._loop = loop

    def close(self):
        for connection in self._connections:
            connection.close()
        ret = asyncio.Future(loop=self._loop)
        ret.set_result(None)
        return ret

    def detach(self, connection):
        self._connections.remove(connection)

    @property
    def connections(self):
        return list(self._connections)

    @property
    def dead_timeout(self):
        return self._dead_timeout

    @property
    def timeout_cutoff(self):
        return self._timeout_cutoff

    @asyncio.coroutine
    def mark_dead(self, connection):
        """
        Mark the connection as dead (failed). Remove it from the live pool and
        put it on a timeout.

        :arg connection: the failed instance
        """
        now = time.monotonic()
        try:
            self._connections.remove(connection)
        except ValueError:
            # connection not alive or another thread marked it already, ignore
            return
        else:
            self._dead_count[connection] += 1
            dead_count = self._dead_count[connection]
            timeout = self._dead_timeout * 2 ** min(dead_count - 1,
                                                    self._timeout_cutoff)
            yield from self._dead.put((now + timeout, connection))
            logger.warning(
                "Connection %r has failed for %i times in a row, "
                "putting on %i second timeout.",
                connection, dead_count, timeout
            )

    @asyncio.coroutine
    def mark_live(self, connection):
        """
        Mark connection as healthy after a resurrection. Resets the fail
        counter for the connection.

        :arg connection: the connection to redeem
        """
        del self._dead_count[connection]

    @asyncio.coroutine
    def resurrect(self, force=False):
        """
        Attempt to resurrect a connection from the dead pool. It will try to
        locate one (not all) eligible (it's timeout is over) connection to
        return to th live pool.

        :arg force: resurrect a connection even if there is none eligible (used
            when we have no live connections)
        """
        if self._dead.empty():
            return

        timeout, connection = self._dead.get_nowait()

        if not force and timeout > time.monotonic():
            # return it back if not eligible and not forced
            yield from self._dead.put((timeout, connection))
            return

        # either we were forced or the connection is elligible to be retried
        self._connections.append(connection)
        logger.info('Resurrecting connection %r (force=%s).',
                    connection, force)

    @asyncio.coroutine
    def get_connection(self):
        """
        Return a connection from the pool using the `ConnectionSelector`
        instance.

        It tries to resurrect eligible connections, forces a resurrection when
        no connections are availible and passes the list of live connections to
        the selector instance to choose from.

        Returns a connection instance
        """
        yield from self.resurrect()

        # no live nodes, resurrect one by force
        if not self._connections:
            yield from self.resurrect(True)

        connection = self._selector.select(self._connections)

        return connection



import asyncio

from .utils import NamespacedClient
from .utils import _make_path
from aioes.exception import NotFoundError

default = object()


class IndicesClient(NamespacedClient):

    @asyncio.coroutine
    def analyze(self, index=None, body=None, *,
                analyzer=default, char_filters=default, field=default,
                filters=default, prefer_local=default, text=default,
                tokenizer=default, token_filters=default):
        """Run analyze tool.

        Perform the analysis process on a text and return the tokens
        breakdown of the text.

        """
        params = {}
        if analyzer is not default:
            params['analyzer'] = analyzer
        if char_filters is not default:
            params['char_filters'] = char_filters
        if field is not default:
            params['field'] = field
        if filters is not default:
            params['filters'] = filters
        if prefer_local is not default:
            params['prefer_local'] = prefer_local
        if text is not default:
            params['text'] = text
        if tokenizer is not default:
            params['tokenizer'] = tokenizer
        if token_filters is not default:
            params['token_filters'] = token_filters

        _, data = yield from self.transport.perform_request(
            'GET',
            _make_path(index, '_analyze'),
            params=params, body=body)
        return data

    @asyncio.coroutine
    def refresh(self, index=None, *,
                allow_no_indices=default, expand_wildcards=default,
                ignore_indices=default, ignore_unavailable=default,
                force=default):
        """Refresh index.

        Explicitly refresh one or more index, making all operations performed
        since the last refresh available for search.
        """
        params = {}
        if force is not default:
            params['force'] = bool(force)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_indices is not default:
            params['ignore_indices'] = ignore_indices
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)

        _, data = yield from self.transport.perform_request(
            'POST',
            _make_path(index, '_refresh'),
            params=params)
        return data

    @asyncio.coroutine
    def flush(self, index=None, *,
              force=default, full=default, allow_no_indices=default,
              expand_wildcards=default, ignore_indices=default,
              ignore_unavailable=default):
        """Explicitly flush one or more indices."""
        params = {}
        if force is not default:
            params['force'] = bool(force)
        if full is not default:
            params['full'] = bool(full)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_indices is not default:
            params['ignore_indices'] = ignore_indices
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)

        _, data = yield from self.transport.perform_request(
            'POST',
            _make_path(index, '_flush'),
            params=params)
        return data

    @asyncio.coroutine
    def create(self, index, body=None, *, timeout=default,
               master_timeout=default):
        """Create an index in Elasticsearch."""
        params = {}
        if timeout is not default:
            params['timeout'] = timeout
        if master_timeout is not default:
            params['master_timeout'] = master_timeout

        _, data = yield from self.transport.perform_request(
            'PUT',
            _make_path(index),
            params=params,
            body=body)
        return data

    @asyncio.coroutine
    def open(self, index, *, timeout=default, master_timeout=default,
             allow_no_indices=default, expand_wildcards=default,
             ignore_unavailable=default):
        """Open a closed index to make it available for search."""
        params = {}
        if timeout is not default:
            params['timeout'] = timeout
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)

        _, data = yield from self.transport.perform_request(
            'POST',
            _make_path(index, '_open'),
            params=params)
        return data

    @asyncio.coroutine
    def close(self, index, *, allow_no_indices=default,
              expand_wildcards=default, ignore_unavailable=default,
              master_timeout=default, timeout=default):
        """Close index.

        Close an index to remove it's overhead from the cluster. Closed index
        is blocked for read/write operations.
        """
        params = {}
        if timeout is not default:
            params['timeout'] = timeout
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)

        _, data = yield from self.transport.perform_request(
            'POST',
            _make_path(index, '_close'),
            params=params)
        return data

    @asyncio.coroutine
    def delete(self, index, *,
               timeout=default, master_timeout=default):
        """Delete an index in Elasticsearch."""
        params = {}
        if timeout is not default:
            params['timeout'] = timeout
        if master_timeout is not default:
            params['master_timeout'] = master_timeout

        _, data = yield from self.transport.perform_request(
            'DELETE',
            _make_path(index),
            params=params)
        return data

    @asyncio.coroutine
    def exists(self, index, *,
               allow_no_indices=default, expand_wildcards=default,
               ignore_unavailable=default, local=default):
        """Return a boolean indicating whether given index exists."""
        params = {}
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if local is not default:
            params['local'] = bool(local)

        try:
            yield from self.transport.perform_request(
                'HEAD', _make_path(index), params=params)
        except NotFoundError:
            return False
        return True

    @asyncio.coroutine
    def exists_type(self, index, doc_type, *,
                    allow_no_indices=default, expand_wildcards=default,
                    ignore_indices=default, ignore_unavailable=default,
                    local=default):
        """Check if a type/types exists in an index/indices."""
        params = {}
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if ignore_indices is not default:
            params['ignore_indices'] = ignore_indices
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if local is not default:
            params['local'] = bool(local)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        try:
            yield from self.transport.perform_request(
                'HEAD', _make_path(index, doc_type), params=params)
        except NotFoundError:
            return False
        return True

    @asyncio.coroutine
    def put_mapping(self, index, doc_type, body, *,
                    allow_no_indices=default, expand_wildcards=default,
                    ignore_conflicts=default, ignore_unavailable=default,
                    master_timeout=default, timeout=default):
        """Register specific mapping definition for a specific type."""
        params = {}
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            params['expand_wildcards'] = bool(expand_wildcards)
        if ignore_conflicts is not default:
            params['ignore_conflicts'] = bool(ignore_conflicts)
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if timeout is not default:
            params['timeout'] = timeout

        _, data = yield from self.transport.perform_request(
            'PUT', _make_path(index, '_mapping', doc_type),
            params=params, body=body)
        return data

    @asyncio.coroutine
    def get_mapping(self, index, doc_type=None, *,
                    ignore_unavailable=default, allow_no_indices=default,
                    expand_wildcards=default, local=default):
        """Retrieve mapping definition of index or index/type."""
        params = {}
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            params['expand_wildcards'] = bool(expand_wildcards)
        if local is not default:
            params['local'] = bool(local)
        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, '_mapping', doc_type),
            params=params
        )
        return data

    @asyncio.coroutine
    def delete_mapping(self, index, doc_type, *,
                       master_timeout=default):
        """Delete a mapping (type) along with its data."""
        params = {}
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        _, data = yield from self.transport.perform_request(
            'DELETE', _make_path(index, '_mapping', doc_type),
            params=params
        )
        return data

    @asyncio.coroutine
    def get_field_mapping(self, field, index=None, doc_type=None, *,
                          include_defaults=default, ignore_unavailable=default,
                          allow_no_indices=default, expand_wildcards=default,
                          local=default):
        """
        Retrieve mapping definition of a specific field.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-get-field-mapping.html>`_

        :arg index: A comma-separated list of index names; use `_all` or empty
            string for all indices
        :arg doc_type: A comma-separated list of document types
        :arg field: A comma-separated list of fields to retrieve the
            mapping for
        :arg include_defaults: A boolean indicating whether to return
            default values
        :arg allow_no_indices: Whether to ignore if a wildcard indices
            expression resolves into no concrete indices. (This includes
            `_all` string or when no indices have been specified)
        :arg expand_wildcards: Whether to expand wildcard expression to
            concrete indices that are open, closed or both.
        :arg ignore_unavailable: Whether specified concrete indices should
            be ignored when unavailable (missing or closed)
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        """
        params = {}
        if include_defaults is not default:
            params['include_defaults'] = bool(include_defaults)
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            params['expand_wildcards'] = bool(expand_wildcards)
        if local is not default:
            params['local'] = bool(local)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, '_mapping', doc_type, 'field', field),
            params=params
        )
        return data

    @asyncio.coroutine
    def put_alias(self, name, index=None, body=None, *,
                  timeout=default, master_timeout=default):
        """
        Create an alias for a specific index/indices.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-aliases.html>`_

        :arg index: A comma-separated list of index names the alias should
            point to (supports wildcards); use `_all` or omit to perform the
            operation on all indices.
        :arg name: The name of the alias to be created or updated
        :arg body: The settings for the alias, such as `routing` or `filter`
        :arg master_timeout: Specify timeout for connection to master
        :arg timeout: Explicit timestamp for the document
        """
        params = {}
        if timeout is not default:
            params['timeout'] = timeout
        if master_timeout is not default:
            params['master_timeout'] = master_timeout

        _, data = yield from self.transport.perform_request(
            'PUT', _make_path(index, '_alias', name),
            params=params, body=body
        )
        return data

    @asyncio.coroutine
    def exists_alias(self, name, index=None, *, allow_no_indices=default,
                     expand_wildcards=default, ignore_indices=default,
                     ignore_unavailable=default, local=default):
        """
        Return a boolean indicating whether given alias exists.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-aliases.html>`_

        :arg name: A comma-separated list of alias names to return
        :arg index: A comma-separated list of index names to filter aliases
        :arg allow_no_indices: Whether to ignore if a wildcard indices
            expression resolves into no concrete indices. (This includes
            `_all` string or when no indices have been specified)
        :arg expand_wildcards: Whether to expand wildcard expression
            to concrete indices that are open, closed or both.
        :arg ignore_indices: When performed on multiple indices, allows to
            ignore `missing` ones (default: none)
        :arg ignore_unavailable: Whether specified concrete indices should
            be ignored when unavailable (missing or closed)
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        """
        params = {}
        if ignore_indices is not default:
            params['ignore_indices'] = bool(ignore_indices)
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            params['expand_wildcards'] = bool(expand_wildcards)
        if local is not default:
            params['local'] = bool(local)

        try:
            yield from self.transport.perform_request(
                'HEAD', _make_path(index, '_alias', name),
                params=params
            )
        except NotFoundError:
            return False
        return True

    @asyncio.coroutine
    def get_alias(self, index=None, name=None, *, allow_no_indices=default,
                  expand_wildcards=default, ignore_indices=default,
                  ignore_unavailable=default, local=default):
        """
        Retrieve a specified alias.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-aliases.html>`_

        :arg name: A comma-separated list of alias names to return
        :arg index: A comma-separated list of index names to filter aliases
        :arg allow_no_indices: Whether to ignore if a wildcard indices
            expression resolves into no concrete indices. (This includes
            `_all` string or when no indices have been specified)
        :arg expand_wildcards: Whether to expand wildcard expression
            to concrete indices that are open, closed or both.
        :arg ignore_indices: When performed on multiple indices, allows to
            ignore `missing` ones, default u'none'
        :arg ignore_unavailable: Whether specified concrete indices should
            be ignored when unavailable (missing or closed)
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        """
        params = {}
        if ignore_indices is not default:
            params['ignore_indices'] = bool(ignore_indices)
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            params['expand_wildcards'] = bool(expand_wildcards)
        if local is not default:
            params['local'] = bool(local)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, '_alias', name),
            params=params
        )
        return data

    @asyncio.coroutine
    def get_aliases(self, index=None, name=None, *, local=default,
                    timeout=default):
        """
        Retrieve specified aliases
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-aliases.html>`_

        :arg index: A comma-separated list of index names to filter aliases
        :arg name: A comma-separated list of alias names to filter
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg timeout: Explicit operation timeout
        """
        params = {}
        if timeout is not default:
            params['timeout'] = timeout
        if local is not default:
            params['local'] = bool(local)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, '_aliases', name),
            params=params
        )
        return data

    @asyncio.coroutine
    def update_aliases(self, body, *, timeout=default,
                       master_timeout=default):
        """
        Update specified aliases.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-aliases.html>`_

        :arg body: The definition of `actions` to perform
        :arg master_timeout: Specify timeout for connection to master
        :arg timeout: Request timeout
        """
        params = {}
        if timeout is not default:
            params['timeout'] = timeout
        if master_timeout is not default:
            params['master_timeout'] = bool(master_timeout)

        _, data = yield from self.transport.perform_request(
            'POST', '/_aliases',
            params=params, body=body
        )
        return data

    @asyncio.coroutine
    def delete_alias(self, index, name, *, timeout=default,
                     master_timeout=default):
        """
        Delete specific alias.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-aliases.html>`_

        :arg index: A comma-separated list of index names (supports wildcards);
            use `_all` for all indices
        :arg name: A comma-separated list of aliases to delete (supports
            wildcards); use `_all` to delete all aliases for the
            specified indices.
        :arg master_timeout: Specify timeout for connection to master
        :arg timeout: Explicit timestamp for the document
        """
        params = {}
        if timeout is not default:
            params['timeout'] = timeout
        if master_timeout is not default:
            params['master_timeout'] = bool(master_timeout)

        _, data = yield from self.transport.perform_request(
            'DELETE', _make_path(index, '_alias', name),
            params=params
        )
        return data

    @asyncio.coroutine
    def put_template(self, name, body, *, create=default, order=default,
                     timeout=default, master_timeout=default,
                     flat_settings=default):
        """
        Create an index template that will automatically be applied to new
        indices created.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-templates.html>`_

        :arg name: The name of the template
        :arg body: The template definition
        :arg create: Whether the index template should only be added if new or
            can also replace an existing one
        :arg order: The order for this template when merging multiple matching
            ones (higher numbers are merged later, overriding the
            lower numbers)
        :arg master_timeout: Specify timeout for connection to master
        :arg timeout: Explicit operation timeout
        :arg flat_settings: Return settings in flat format (default: false)
        """
        params = {}

        if create is not default:
            params['create'] = create
        if order is not default:
            params['order'] = order
        if timeout is not default:
            params['timeout'] = timeout
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if flat_settings is not default:
            params['flat_settings'] = bool(flat_settings)

        _, data = yield from self.transport.perform_request(
            'PUT', _make_path('_template', name),
            params=params, body=body
        )
        return data

    @asyncio.coroutine
    def exists_template(self, name, *, local=default):
        """
        Return a boolean indicating whether given template exists.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-templates.html>`_

        :arg name: The name of the template
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        """
        params = {}
        if local is not default:
            params['local'] = bool(local)

        try:
            yield from self.transport.perform_request(
                'HEAD', _make_path('_template', name),
                params=params
            )
        except NotFoundError:
            return False
        return True

    @asyncio.coroutine
    def get_template(self, name=None, *, flat_settings=default,
                     local=default):
        """
        Retrieve an index template by its name.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-templates.html>`_

        :arg name: The name of the template
        :arg flat_settings: Return settings in flat format (default: false)
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        """
        params = {}
        if local is not default:
            params['local'] = bool(local)
        if flat_settings is not default:
            params['flat_settings'] = bool(flat_settings)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_template', name),
            params=params
        )
        return data

    @asyncio.coroutine
    def delete_template(self, name, *, timeout=default,
                        master_timeout=default):
        """
        Delete an index template by its name.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-templates.html>`_

        :arg name: The name of the template
        :arg master_timeout: Specify timeout for connection to master
        :arg timeout: Explicit operation timeout
        """
        params = {}
        if timeout is not default:
            params['timeout'] = timeout
        if master_timeout is not default:
            params['master_timeout'] = master_timeout

        _, data = yield from self.transport.perform_request(
            'DELETE', _make_path('_template', name),
            params=params
        )
        return data

    @asyncio.coroutine
    def get_settings(self, index=None, name=None, *,
                     expand_wildcards=default, ignore_indices=default,
                     ignore_unavailable=default, flat_settings=default,
                     local=default):
        """Retrieve settings for one or more (or all) indices."""
        params = {}
        if ignore_indices is not default:
            params['ignore_indices'] = str(ignore_indices)
        if flat_settings is not default:
            params['flat_settings'] = bool(flat_settings)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if local is not default:
            params['local'] = bool(local)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, '_settings', name),
            params=params)
        return data

    @asyncio.coroutine
    def put_settings(self, body, index=None, *,
                     allow_no_indices=default, expand_wildcards=default,
                     flat_settings=default, ignore_unavailable=default,
                     master_timeout=default):
        """Change specific index level settings in real time."""
        params = {}
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if flat_settings is not default:
            params['flat_settings'] = bool(flat_settings)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout

        _, data = yield from self.transport.perform_request(
            'PUT', _make_path(index, '_settings'),
            params=params, body=body)
        return data

    @asyncio.coroutine
    def put_warmer(self, name, body, index=None, doc_type=None, *,
                   allow_no_indices=default, expand_wildcards=default,
                   ignore_unavailable=default, master_timeout=default):
        """
        Create an index warmer to run registered search requests to warm up the
        index before it is available for search.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-warmers.html>`_

        :arg name: The name of the warmer
        :arg body: The search request definition for the warmer
            (query, filters, facets, sorting, etc)
        :arg index: A comma-separated list of index names to register
            the warmer for; use `_all` or omit to perform the operation
            on all indices
        :arg doc_type: A comma-separated list of document types to register the
            warmer for; leave empty to perform the operation on all types
        :arg allow_no_indices: Whether to ignore if a wildcard indices
            expression resolves into no concrete indices in the search request
            to warm. (This includes `_all` string or when no indices have been
            specified)
        :arg expand_wildcards: Whether to expand wildcard expression
            to concrete indices that are open, closed or both, in the
            search request to warm., default u'open'
        :arg ignore_unavailable: Whether specified concrete indices should be
            ignored when unavailable (missing or closed) in the search request
            to warm
        :arg master_timeout: Specify timeout for connection to master
        """
        params = {}

        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            params['expand_wildcards'] = bool(expand_wildcards)

        if doc_type and not index:
            index = '_all'
        _, data = yield from self.transport.perform_request(
            'PUT', _make_path(index, doc_type, '_warmer', name),
            params=params, body=body
        )
        return data

    @asyncio.coroutine
    def get_warmer(self, index=None, doc_type=None, name=None, *,
                   allow_no_indices=default, expand_wildcards=default,
                   ignore_unavailable=default, local=default):
        """
        Retreieve an index warmer.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-warmers.html>`_

        :arg index: A comma-separated list of index names to restrict the
            operation; use `_all` to perform the operation on all indices
        :arg doc_type: A comma-separated list of document types to restrict the
            operation; leave empty to perform the operation on all types
        :arg name: The name of the warmer (supports wildcards); leave empty to
            get all warmers
        :arg allow_no_indices: Whether to ignore if a wildcard indices
            expression resolves into no concrete indices. (This includes `_all`
            string or when no indices have been specified)
        :arg expand_wildcards: Whether to expand wildcard expression
            to concrete indices that are open, closed or both. default u'open'
        :arg ignore_unavailable: Whether specified concrete indices should be
            ignored when unavailable (missing or closed)
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        """
        params = {}

        if local is not default:
            params['local'] = bool(local)
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            params['expand_wildcards'] = bool(expand_wildcards)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, doc_type, '_warmer', name),
            params=params
        )
        return data

    @asyncio.coroutine
    def delete_warmer(self, index, name, *, master_timeout=default):
        """
        Delete an index warmer.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-warmers.html>`_

        :arg index: A comma-separated list of index names to delete
            warmers from (supports wildcards); use `_all` to perform
            the operation on all indices.
        :arg name: A comma-separated list of warmer names to delete (supports
            wildcards); use `_all` to delete all warmers in the
            specified indices.
        :arg master_timeout: Specify timeout for connection to master
        """
        params = {}

        if master_timeout is not default:
            params['master_timeout'] = master_timeout

        _, data = yield from self.transport.perform_request(
            'DELETE', _make_path(index, '_warmer', name),
            params=params
        )
        return data

    @asyncio.coroutine
    def snapshot_index(self, index=None, *, allow_no_indices=default,
                       expand_wildcards=default, ignore_indices=default,
                       ignore_unavailable=default):
        """
        Explicitly perform a snapshot through the gateway of one or more
        indices (backup them).
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/indices-gateway-snapshot.html>`_

        :arg index: A comma-separated list of index names; use `_all` or empty
            string for all indices
        :arg allow_no_indices: Whether to ignore if a wildcard indices
            expression resolves into no concrete indices. (This includes
            `_all` string or when no indices have been specified)
        :arg expand_wildcards: Whether to expand wildcard expression
            to concrete indices that are open, closed or both.
        :arg ignore_indices: When performed on multiple indices, allows to
            ignore `missing` ones (default: none)
        :arg ignore_unavailable: Whether specified concrete indices should
            be ignored when unavailable (missing or closed)
        """
        params = {}

        if ignore_indices is not default:
            params['ignore_indices'] = bool(ignore_unavailable)
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            params['expand_wildcards'] = bool(expand_wildcards)

        _, data = yield from self.transport.perform_request(
            'POST',
            _make_path(index, '_gateway', 'snapshot'), params=params
        )
        return data

    @asyncio.coroutine
    def status(self, index=None, *,
               allow_no_indices=default, expand_wildcards=default,
               ignore_indices=default, ignore_unavailable=default,
               operation_threading=default, recovery=default, snapshot=default,
               human=default):
        """Get a comprehensive status information of one or more indices."""
        params = {}
        if ignore_indices is not default:
            params['ignore_indices'] = ignore_indices
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if recovery is not default:
            params['recovery'] = bool(recovery)
        if snapshot is not default:
            params['snapshot'] = bool(snapshot)
        if operation_threading is not default:
            params['operation_threading'] = operation_threading
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if human is not default:
            params['human'] = bool(human)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, '_status'),
            params=params)
        return data

    @asyncio.coroutine
    def stats(self, index=None, *, metric=default,
              completion_fields=default, docs=default,
              fielddata_fields=default, fields=default,
              groups=default, allow_no_indices=default,
              expand_wildcards=default, ignore_indices=default,
              ignore_unavailable=default, human=default, level=default,
              types=default):
        """Retrieve statistics on operations happening on an index."""
        params = {}
        if completion_fields is not default:
            params['completion_fields'] = completion_fields
        if docs is not default:
            params['docs'] = docs
        if types is not default:
            params['types'] = types
        if fielddata_fields is not default:
            params['fielddata_fields'] = fielddata_fields
        if fields is not default:
            params['fields'] = fields
        if groups is not default:
            params['groups'] = groups

        if ignore_indices is not default:
            params['ignore_indices'] = ignore_indices
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if human is not default:
            params['human'] = bool(human)
        if level is not default:
            if not isinstance(level, str):
                raise TypeError("'level' parameter is not a string")
            elif level.lower() in ('cluster', 'indices', 'shards'):
                params['level'] = level.lower()
            else:
                raise ValueError("'level' parameter should be one"
                                 " of 'cluster', 'indices', 'shards'")
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if metric is not default:
            if not isinstance(metric, str):
                raise TypeError("'metric' parameter is not a string")
            elif metric.lower() in ('_all', 'completion', 'docs', 'fielddata',
                                    'filter_cache', 'flush', 'get', 'id_cache',
                                    'indexing', 'merge', 'percolate',
                                    'refresh', 'search', 'segments', 'store',
                                    'warmer'):
                params['metric'] = metric.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of '_all', 'completion', 'docs', "
                                 "'fielddata', 'filter_cache', 'flush', "
                                 "'get', 'id_cache', 'indexing', 'merge', "
                                 "'percolate', 'refresh', 'search', "
                                 "'segments', 'store', 'warmer'")

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, '_stats', metric),
            params=params)
        return data

    @asyncio.coroutine
    def segments(self, index=None, *,
                 allow_no_indices=default, expand_wildcards=default,
                 ignore_indices=default, ignore_unavailable=default,
                 human=default):
        """Get segments information.

        Provide low level segments information that a Lucene index (shard
        level) is built with.
        """
        params = {}
        if ignore_indices is not default:
            params['ignore_indices'] = ignore_indices
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)
        if human is not default:
            params['human'] = bool(human)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, '_segments'), params=params)
        return data

    @asyncio.coroutine
    def optimize(self, index=None, *,
                 flush=default, allow_no_indices=default,
                 expand_wildcards=default, ignore_indices=default,
                 ignore_unavailable=default, max_num_segments=default,
                 only_expunge_deletes=default, operation_threading=default,
                 wait_for_merge=default, force=default):
        """Explicitly optimize one or more indices through an API."""
        params = {}
        if force is not default:
            params['force'] = bool(force)
        if flush is not default:
            params['flush'] = bool(flush)
        if max_num_segments is not default:
            params['max_num_segments'] = int(max_num_segments)
        if ignore_indices is not default:
            params['ignore_indices'] = ignore_indices
        if only_expunge_deletes is not default:
            params['only_expunge_deletes'] = bool(only_expunge_deletes)
        if operation_threading is not default:
            params['operation_threading'] = operation_threading
        if wait_for_merge is not default:
            params['wait_for_merge'] = bool(wait_for_merge)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)

        _, data = yield from self.transport.perform_request(
            'POST', _make_path(index, '_optimize'), params=params)
        return data

    @asyncio.coroutine
    def validate_query(self, index=None, doc_type=None, body=None, *,
                       explain=default, allow_no_indices=default,
                       expand_wildcards=default, ignore_indices=default,
                       ignore_unavailable=default, operation_threading=default,
                       q=default, source=default):
        """Validate a potentially expensive query without executing it."""
        params = {}
        if explain is not default:
            params['explain'] = bool(explain)
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if q is not default:
            params['q'] = str(q)
        if ignore_indices is not default:
            params['ignore_indices'] = ignore_indices
        if source is not default:
            params['source'] = str(source)
        if operation_threading is not default:
            params['operation_threading'] = operation_threading
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, doc_type, '_validate', 'query'),
            params=params, body=body)
        return data

    @asyncio.coroutine
    def clear_cache(self, index=None, *,
                    field_data=default, fielddata=default, fields=default,
                    filter=default, filter_cache=default, filter_keys=default,
                    id=default, id_cache=default, allow_no_indices=default,
                    expand_wildcards=default, ignore_indices=default,
                    ignore_unavailable=default, recycler=default):
        """Clear cache.

        Clear either all caches or specific cached associated with one or
        more indices.
        """
        params = {}
        if recycler is not default:
            params['recycler'] = bool(recycler)
        if id_cache is not default:
            params['id_cache'] = bool(id_cache)
        if id is not default:
            params['id'] = bool(id)
        if filter_keys is not default:
            params['filter_keys'] = filter_keys
        if filter_cache is not default:
            params['filter_cache'] = bool(filter_cache)
        if filter is not default:
            params['filter'] = bool(filter)
        if fields is not default:
            params['fields'] = fields
        if field_data is not default:
            params['field_data'] = bool(field_data)
        if fielddata is not default:
            params['fielddata'] = bool(fielddata)
        if ignore_indices is not default:
            params['ignore_indices'] = ignore_indices
        if allow_no_indices is not default:
            params['allow_no_indices'] = bool(allow_no_indices)
        if expand_wildcards is not default:
            if not isinstance(expand_wildcards, str):
                raise TypeError("'expand_wildcards' parameter is not a string")
            elif expand_wildcards.lower() in ('open', 'closed'):
                params['expand_wildcards'] = expand_wildcards.lower()
            else:
                raise ValueError("'expand_wildcards' parameter should be one"
                                 " of 'open', 'closed'")
        if ignore_unavailable is not default:
            params['ignore_unavailable'] = bool(ignore_unavailable)

        _, data = yield from self.transport.perform_request(
            'POST', _make_path(index, '_cache', 'clear'),
            params=params)
        return data

    @asyncio.coroutine
    def recovery(self, index=None, *,
                 active_only=default, detailed=default, human=default):
        """Recover an index.

        The indices recovery API provides insight into on-going shard
        recoveries. Recovery status may be reported for specific indices, or
        cluster-wide.
        """
        params = {}
        if active_only is not default:
            params['active_only'] = bool(active_only)
        if detailed is not default:
            params['detailed'] = bool(detailed)
        if human is not default:
            params['human'] = bool(human)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path(index, '_recovery'), params=params)
        return data



import asyncio
import unittest
from aioes import Elasticsearch
from aioes.exception import NotFoundError
import pprint
pp = pprint.pprint


class TestCluster(unittest.TestCase):
    def setUp(self):
        self._index = 'elastic_search'
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.cl = Elasticsearch([{'host': 'localhost'}], loop=self.loop)
        self.addCleanup(self.cl.close)
        try:
            self.loop.run_until_complete(
                self.cl.delete(self._index, refresh=True))
        except NotFoundError:
            pass

    def tearDown(self):
        self.loop.close()

    def test_health(self):
        @asyncio.coroutine
        def go():
            data = yield from self.cl.cluster.health()
            self.assertIn('status', data)
            data = yield from self.cl.cluster.health(
                level='indices',
                local=True,
                master_timeout='1s',
                timeout='1s',
                wait_for_active_shards=1,
                wait_for_nodes='>2',
                wait_for_relocating_shards=0)
            self.assertIn('status', data)
            with self.assertRaises(TypeError):
                yield from self.cl.cluster.health(level=1)
            with self.assertRaises(ValueError):
                yield from self.cl.cluster.health(level='1')
            with self.assertRaises(TypeError):
                yield from self.cl.cluster.health(wait_for_status=1)
            with self.assertRaises(ValueError):
                yield from self.cl.cluster.health(wait_for_status='1')

        self.loop.run_until_complete(go())

    def test_pending_tasks(self):
        @asyncio.coroutine
        def go():
            data = yield from self.cl.cluster.pending_tasks()
            self.assertIn('tasks', data)

        self.loop.run_until_complete(go())

    def test_state(self):
        @asyncio.coroutine
        def go():
            data = yield from self.cl.cluster.state()
            self.assertIn('routing_nodes', data)
            self.assertIn('master_node', data)

            # create index
            yield from self.cl.create(
                self._index, 'tweet',
                {
                    'user': 'Bob',
                },
                '1'
            )
            data = yield from self.cl.cluster.state(index=self._index)
            self.assertIn('routing_nodes', data)
            self.assertIn('master_node', data)

        self.loop.run_until_complete(go())

    def test_stats(self):
        @asyncio.coroutine
        def go():
            data = yield from self.cl.cluster.stats()
            self.assertIn('indices', data)
            self.assertIn('nodes', data)

        self.loop.run_until_complete(go())

    def test_reroute(self):
        @asyncio.coroutine
        def go():
            # create index
            yield from self.cl.create(
                self._index, 'tweet',
                {
                    'user': 'Bob',
                },
                '1'
            )

            # get node's name
            nodes = yield from self.cl.nodes.info()
            node = list(nodes['nodes'].keys())[0]

            b = {
                "commands": [
                    {
                        "cancel": {
                            "index": self._index,
                            "shard": 1,
                            "node": node,
                            'allow_primary': True,
                        }
                    }
                ]
            }
            data = yield from self.cl.cluster.reroute(body=b, dry_run=True)
            self.assertTrue(data['acknowledged'])

        self.loop.run_until_complete(go())

    def test_get_settings(self):
        @asyncio.coroutine
        def go():
            data = yield from self.cl.cluster.get_settings()
            self.assertIn('persistent', data)
            self.assertIn('transient', data)

        self.loop.run_until_complete(go())

    def test_put_settings(self):
        @asyncio.coroutine
        def go():
            b = {
                "transient": {
                    "cluster.routing.allocation.enable": "all"
                }
            }
            data = yield from self.cl.cluster.put_settings(b)
            routing_settings = data['transient']['cluster']['routing']
            self.assertEqual(
                routing_settings['allocation']['enable'],
                'all'
            )

        self.loop.run_until_complete(go())



import asyncio
import collections
import itertools
import json
import random
import re
import time
import urllib.parse

from .connection import Connection
from .exception import ConnectionError, TransportError
from .pool import ConnectionPool

Endpoint = collections.namedtuple('TCPEndpoint', 'scheme host port')


def validate_endpoint(endpoint):
    if not isinstance(endpoint.port, int):
        raise ValueError('bad port {}'.format(endpoint.port))
    if endpoint.scheme not in ('http', 'https'):
        raise ValueError('bad scheme {}'.format(endpoint.scheme))
    if not isinstance(endpoint.host, str) or not endpoint.host:
        raise ValueError('bad host {}'.format(endpoint.host))


def get_default_port(scheme):
    return 443 if scheme == 'https' else 9200


DEFAULT_SCHEME = 'http'


class Transport:
    """Encapsulation of transport-related to logic.

    Handles instantiation of the individual connections as well as
    creating a connection pool to hold them.

    Main interface is the `perform_request` method.

    """

    # get ip/port from "inet[wind/127.0.0.1:9200]" or "127.0.0.1:9200"
    ADDRESS_RE = re.compile(
            r'(?:^|/)(?P<host>[\.:0-9a-f]*):(?P<port>[0-9]+)\]?$')

    def __init__(self, endpoints, *,
                 sniffer_interval=None, sniffer_timeout=0.1, max_retries=3,
                 loop, verify_ssl=True):
        self._loop = loop
        self._endpoints = self._convert_endpoints(endpoints)
        self._pool = ConnectionPool([], loop=loop)
        self._verify_ssl = verify_ssl
        self._reinitialize_endpoints()
        self._seed_connections = list(self._pool.connections)
        self._sniffer_interval = sniffer_interval
        self._sniffer_timeout = sniffer_timeout
        self._last_sniff = time.monotonic()
        self._max_retries = max_retries

    @property
    def max_retries(self):
        return self._max_retries

    @property
    def last_sniff(self):
        return self._last_sniff

    @property
    def sniffer_interval(self):
        return self._sniffer_interval

    @property
    def sniffer_timeout(self):
        return self._sniffer_timeout

    @property
    def endpoints(self):
        return list(self._endpoints)

    @endpoints.setter
    def endpoints(self, endpoints):
        self._endpoints = self._convert_endpoints(endpoints)
        self._reinitialize_endpoints()

    def close(self):
        return self._pool.close()

    def _convert_endpoints(self, endpoints):
        ret = []
        for e in endpoints:
            if isinstance(e, Endpoint):
                endpoint = e
            elif isinstance(e, dict):
                try:
                    host = e['host']
                except KeyError:
                    raise RuntimeError("Bad endpoint {}".format(e))
                port = e.get('port')
                scheme = e.get('scheme', DEFAULT_SCHEME)
                if port is None:
                    port = get_default_port(scheme)
                endpoint = Endpoint(scheme, host, port)
            elif isinstance(e, str):
                if not re.match(r'^(\w*\:?)//.*', e):
                    e = '{}://{}'.format(DEFAULT_SCHEME, e)
                parts = urllib.parse.urlparse(e)
                if parts.scheme:
                    scheme = parts.scheme
                else:
                    scheme = DEFAULT_SCHEME
                try:
                    port = parts.port
                except ValueError:
                    raise RuntimeError("Bad endpoint {}".format(e))
                if port is None:
                    port = get_default_port(scheme)
                if parts.hostname:
                    auth = ':'.join(
                        [x for x in (parts.username, parts.password) if x]
                    )
                    if auth:
                        host = '{}@{}'.format(auth, parts.hostname)
                    else:
                        host = parts.hostname
                endpoint = Endpoint(scheme, host, port)
            else:
                raise RuntimeError("Bad endpoint {}".format(e))
            try:
                validate_endpoint(endpoint)
            except ValueError:
                raise RuntimeError("Bad endpoint {}".format(endpoint))
            ret.append(endpoint)
        return ret

    def _reinitialize_endpoints(self):
        old_connections = {c.endpoint: c for c in self._pool.connections}
        connections = []
        for endpoint in self._endpoints:
            if endpoint in old_connections:
                connection = old_connections[endpoint]
                connections.append(connection)
                self._pool.detach(connection)
            else:
                connections.append(Connection(
                    endpoint,
                    loop=self._loop,
                    verify_ssl=self._verify_ssl))
        self._pool.close()
        random.shuffle(connections)
        self._pool = ConnectionPool(connections, loop=self._loop)

    @asyncio.coroutine
    def get_connection(self):
        """
        Retreive a :class:`~aioes.Connection` instance from the
        :class:`~aioes.ConnectionPool` instance.
        """
        if self._sniffer_interval:
            if time.monotonic() >= self._last_sniff + self._sniffer_interval:
                yield from self.sniff_endpoints()
        ret = yield from self._pool.get_connection()
        return ret

    @asyncio.coroutine
    def sniff_endpoints(self):
        """Obtain a list of nodes from the cluster and create a new connection
        pool using the information retrieved.

        To extract the node connection parameters use the
        `nodes_to_endpoint_callback`.

        """
        previous_sniff = self._last_sniff
        try:
            # reset last_sniff timestamp
            self._last_sniff = time.monotonic()
            # go through all current connections as well as the
            # seed_connections for good measure
            for c in itertools.chain(self._pool.connections,
                                     self._seed_connections):
                try:
                    # use small timeout for the sniffing request,
                    # should be a fast api call
                    _, headers, node_info = yield from asyncio.wait_for(
                        c.perform_request(
                            'GET',
                            '/_nodes/_all/clear',
                            None,
                            None),
                        timeout=self._sniffer_timeout,
                        loop=self._loop)
                except ConnectionError:
                    continue
                try:
                    node_info = json.loads(node_info)
                except (TypeError, ValueError):
                    continue
                break
            else:
                raise TransportError("N/A", "Unable to sniff endpoints.")
        except:
            # keep the previous value on error
            self._last_sniff = previous_sniff
            raise

        endpoints = []
        address = 'http_address'
        for n in node_info['nodes'].values():
            match = self.ADDRESS_RE.search(n.get(address, ''))
            if not match:
                continue

            dct = match.groupdict()
            host = dct['host']
            if 'port' in dct:
                port = int(dct['port'])
            else:
                port = 9200
            scheme = dct.get('scheme') or DEFAULT_SCHEME
            attrs = n.get('attributes', {})
            if not (attrs.get('data', 'true') == 'false' and
                    attrs.get('client', 'false') == 'false' and
                    attrs.get('master', 'true') == 'true'):
                endpoints.append(Endpoint(scheme, host, port))

        # we weren't able to get any nodes, maybe using an incompatible
        # transport_schema or host_info_callback blocked all - raise error.
        if not endpoints:
            raise TransportError(
                "N/A",
                "Unable to sniff endpoints - no viable endpoints found.")

        self.endpoints = endpoints

    @asyncio.coroutine
    def _mark_dead(self, connection):
        """
        Mark a connection as dead (failed) in the connection pool. If sniffing
        on failure is enabled this will initiate the sniffing process.

        :arg connection: instance of :class:`~aioes.Connection` that failed
        """
        # mark as dead even when sniffing to avoid hitting this endpoint
        # during the sniff process

        yield from self._pool.mark_dead(connection)
        yield from self.sniff_endpoints()

    @asyncio.coroutine
    def perform_request(self, method, url, params=None, body=None,
                        *, request_timeout=None, decoder=json.loads):
        """
        Perform the actual request. Retrieve a connection from the connection
        pool, pass all the information to it's perform_request method and
        return the data.

        If an exception was raised, mark the connection as failed and retry (up
        to `max_retries` times).

        If the operation was succesful and the connection used was previously
        marked as dead, mark it as live, resetting it's failure count.

        :arg method: HTTP method to use
        :arg url: absolute url (without endpoint) to target
        :arg params: dictionary of query parameters, will be handed over to the
          underlying :class:`~elasticsearch.Connection` class for serialization
        :arg body: body of the request, will be serializes using serializer and
            passed to the connection
        """
        if body is not None:
            if not isinstance(body, (str, bytes)):
                body = json.dumps(body)

            if not isinstance(body, bytes):
                body = body.encode('utf-8')

        for attempt in range(self.max_retries + 1):
            connection = yield from self.get_connection()

            try:
                status, headers, data = yield from asyncio.wait_for(
                    connection.perform_request(
                        method,
                        url,
                        params,
                        body),
                    request_timeout,
                    loop=self._loop)
            except ConnectionError:
                yield from self._mark_dead(connection)

                # raise exception on last retry
                if attempt == self.max_retries:
                    raise
            else:
                # connection didn't fail, confirm it's live status
                yield from self._pool.mark_live(connection)
                if data:
                    data = decoder(data)
                return status, data



import asyncio
import json
import logging

import aiohttp
from .exception import HTTP_EXCEPTIONS, TransportError

logger = logging.getLogger(__name__)


class Connection:
    """
    Class responsible for maintaining a connection to an Elasticsearch node.
    Holds persistent connection pool to it.

    Also responsible for logging.
    """

    def __init__(self, endpoint, *, loop, verify_ssl=True):
        self._endpoint = endpoint
        self._session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(
                use_dns_cache=True,
                loop=loop,
                verify_ssl=verify_ssl),
            loop=loop)
        self._base_url = '{0.scheme}://{0.host}:{0.port}/'.format(endpoint)

    @property
    def endpoint(self):
        return self._endpoint

    def close(self):
        return self._session.close()

    @asyncio.coroutine
    def perform_request(self, method, url, params, body):
        url = self._base_url + url
        resp = yield from self._session.request(
            method, url, params=params, data=body)
        resp_body = yield from resp.text()
        if not (200 <= resp.status <= 300):
            extra = None
            try:
                extra = json.loads(resp_body)
            except ValueError:
                pass
            exc_class = HTTP_EXCEPTIONS.get(resp.status, TransportError)
            raise exc_class(resp.status, resp_body, extra)
        return resp.status, resp.headers, resp_body



import asyncio

from .utils import NamespacedClient
from .utils import _make_path

default = object()


class SnapshotClient(NamespacedClient):

    @asyncio.coroutine
    def create(self, repository, snapshot, body=None, *,
               master_timeout=default, wait_for_completion=default):
        """
        Create a snapshot in repository
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/modules-snapshots.html>`_

        :arg repository: A repository name
        :arg snapshot: A snapshot name
        :arg body: The snapshot definition
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        :arg wait_for_completion: Should this request wait until
            the operation has completed before returning, default False
        """
        params = {}

        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if wait_for_completion is not default:
            params['wait_for_completion'] = bool(wait_for_completion)

        _, data = yield from self.transport.perform_request(
            'PUT', _make_path('_snapshot', repository, snapshot),
            params=params, body=body
        )
        return data

    @asyncio.coroutine
    def delete(self, repository, snapshot, *, master_timeout=default):
        """
        Deletes a snapshot from a repository.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/modules-snapshots.html>`_

        :arg repository: A repository name
        :arg snapshot: A snapshot name
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        """
        params = {}

        if master_timeout is not default:
            params['master_timeout'] = master_timeout

        _, data = yield from self.transport.perform_request(
            'DELETE',
            _make_path('_snapshot', repository, snapshot),
            params=params
        )
        return data

    @asyncio.coroutine
    def get(self, repository, snapshot, *, master_timeout=default):
        """
        Retrieve information about a snapshot.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/modules-snapshots.html>`_

        :arg repository: A comma-separated list of repository names
        :arg snapshot: A comma-separated list of snapshot names
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        """
        params = {}

        if master_timeout is not default:
            params['master_timeout'] = master_timeout

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_snapshot', repository, snapshot),
            params=params
        )
        return data

    @asyncio.coroutine
    def delete_repository(self, repository, *, master_timeout=default,
                          timeout=default):
        """
        Removes a shared file system repository.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/modules-snapshots.html>`_

        :arg repository: A comma-separated list of repository names
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        :arg timeout: Explicit operation timeout
        """
        params = {}

        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if timeout is not default:
            params['timeout'] = timeout

        _, data = yield from self.transport.perform_request(
            'DELETE',
            _make_path('_snapshot', repository),
            params=params
        )
        return data

    @asyncio.coroutine
    def get_repository(self, repository=None, *, local=default,
                       master_timeout=default):
        """
        Return information about registered repositories.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/modules-snapshots.html>`_

        :arg repository: A comma-separated list of repository names
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        """
        params = {}

        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if local is not default:
            params['local'] = bool(local)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_snapshot', repository),
            params=params
        )
        return data

    @asyncio.coroutine
    def create_repository(self, repository, body, *, master_timeout=default,
                          timeout=default):
        """
        Registers a shared file system repository.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/modules-snapshots.html>`_

        :arg repository: A repository name
        :arg body: The repository definition
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        :arg timeout: Explicit operation timeout
        """
        params = {}

        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if timeout is not default:
            params['timeout'] = timeout

        _, data = yield from self.transport.perform_request(
            'PUT', _make_path('_snapshot', repository),
            params=params, body=body
        )
        return data

    @asyncio.coroutine
    def restore(self, repository, snapshot, body=None, *,
                master_timeout=default, wait_for_completion=default):
        """
        Restore a snapshot.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/modules-snapshots.html>`_

        :arg repository: A repository name
        :arg snapshot: A snapshot name
        :arg body: Details of what to restore
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        :arg wait_for_completion: Should this request wait until the operation
            has completed before returning, default False
        """
        params = {}

        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if wait_for_completion is not default:
            params['wait_for_completion'] = bool(wait_for_completion)

        _, data = yield from self.transport.perform_request(
            'POST', _make_path('_snapshot', repository, snapshot, '_restore'),
            params=params, body=body
        )
        return data

    @asyncio.coroutine
    def status(self, repository=None, snapshot=None, *,
               master_timeout=default):
        """Get snapshot status

        :arg repository: A repository name
        :arg snapshot: A comma-separated list of snapshot names
        :arg master_timeout: Explicit operation timeout for connection
            to master node
        """
        params = {}
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        _, data = yield from self.transport.perform_request(
            'GET',
            _make_path('_snapshot', repository, snapshot, '_status'),
            params=params)
        return data



import os
import shutil
import random
import asyncio
import tempfile
import unittest
from aioes import Elasticsearch
from aioes.exception import NotFoundError


class TestSnapshot(unittest.TestCase):

    def _create_temp_dir(self):
        '''
        Sice elasticsearch may be launched under the other user,
        tempfile.TemporaryDirectory can't be used, because python
        can't cleanup it after elasticsearch user.
        So we are just leaving it there.
        '''
        characters = "abcdefghijklmnopqrstuvwxyz0123456789_"
        temp_dir = tempfile.gettempdir()
        temp_prefix = '/' + tempfile.gettempprefix()
        temp_name = temp_prefix + ''.join(
            [random.choice(characters) for _ in range(8)]
        )

        dir_path = os.path.join(temp_dir + temp_name)
        os.makedirs(dir_path)
        return dir_path

    def _cleanup_dirs(self, dir_path):
        try:
            shutil.rmtree(dir_path)
        except PermissionError:
            # if subdirs were created by other user
            pass

    def setUp(self):
        self._index = 'elastic_search'
        self.repo_name = 'test_repo'
        self.repo_path = self._create_temp_dir()
        # otherwise elasticsearch can't access it
        os.chmod(self.repo_path, 0o777)

        self.snapshot_name = 'test_snapshot'
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.cl = Elasticsearch([{'host': 'localhost'}], loop=self.loop)
        self.addCleanup(self.cl.close)
        try:
            self.loop.run_until_complete(
                self.cl.delete(self._index, refresh=True))
        except NotFoundError:
            pass

    def tearDown(self):
        self._cleanup_dirs(self.repo_path)
        # cleaning up just in case
        try:
            self.loop.run_until_complete(
                self.cl.snapshot.delete(self.repo_name, self.snapshot_name))
        except NotFoundError:
            pass
        try:
            self.loop.run_until_complete(
                self.cl.snapshot.delete_repository(self.repo_name))
        except NotFoundError:
            pass

        # close loop
        self.loop.close()

    def test_repository(self):
        @asyncio.coroutine
        def go():
            ret = yield from self.cl.snapshot.get_repository()
            self.assertFalse(ret)

            b = {
                "type": "fs",
                "settings": {
                    "location": self.repo_path,
                }
            }
            ret = yield from self.cl.snapshot.create_repository(
                self.repo_name, b
            )
            self.assertTrue(ret['acknowledged'])

            ret = yield from self.cl.snapshot.get_repository(
                self.repo_name
            )
            self.assertEqual(ret[self.repo_name], b)

            ret = yield from self.cl.snapshot.delete_repository(
                self.repo_name
            )
            self.assertTrue(ret['acknowledged'])

        self.loop.run_until_complete(go())

    def test_snapshot(self):
        @asyncio.coroutine
        def go():
            # creating index
            yield from self.cl.create(
                self._index, 'tweet',
                {
                    'user': 'Bob',
                },
                '1'
            )

            # creating repo
            repo_body = {
                "type": "fs",
                "settings": {
                    "location": self.repo_path,
                }
            }

            ret = yield from self.cl.snapshot.create_repository(
                self.repo_name, repo_body
            )
            self.assertTrue(ret['acknowledged'])

            # creating snapshot
            yield from self.cl.snapshot.create(
                self.repo_name, self.snapshot_name,
                wait_for_completion=True
            )

            # checking that snapshot was created
            ret = yield from self.cl.snapshot.get(
                self.repo_name, self.snapshot_name,
            )
            self.assertEqual(
                ret['snapshots'][0]['snapshot'],
                self.snapshot_name
            )
            self.assertEqual(
                ret['snapshots'][0]['state'],
                'SUCCESS'
            )

            # restoring snapshot
            restore_body = {
                "indices": self._index,
            }
            yield from self.cl.indices.close(self._index)
            ret = yield from self.cl.snapshot.restore(
                self.repo_name, self.snapshot_name,
                body=restore_body
            )
            self.assertTrue(ret['accepted'])

            # deleting snapshot
            ret = yield from self.cl.snapshot.delete(
                self.repo_name, self.snapshot_name,
            )
            self.assertTrue(ret['acknowledged'])

            # deleting repo
            ret = yield from self.cl.snapshot.delete_repository(
                self.repo_name
            )
            self.assertTrue(ret['acknowledged'])

        self.loop.run_until_complete(go())

    def test_status(self):
        @asyncio.coroutine
        def go():
            ret = yield from self.cl.snapshot.status()
            self.assertEqual({'snapshots': []}, ret)

        self.loop.run_until_complete(go())



import asyncio
import time
import unittest
import urllib.parse

from aioes.transport import Endpoint, Transport


class TestTransport(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

    def tearDown(self):
        self.loop.close()

    def make_transport(self, endpoints=[{'host': 'localhost'}],
                       sniffer_interval=None):
        tr = Transport(endpoints, loop=self.loop,
                       sniffer_interval=sniffer_interval)
        self.addCleanup(tr.close)
        return tr

    def test_ctor(self):
        tr = self.make_transport()
        self.assertEqual(3, tr.max_retries)
        self.assertGreaterEqual(time.monotonic(), tr.last_sniff)
        self.assertIsNone(tr.sniffer_interval)
        self.assertAlmostEqual(0.1, tr.sniffer_timeout)
        self.assertEqual([Endpoint('http', 'localhost', 9200)], tr.endpoints)
        self.assertEqual(1, len(tr._pool.connections))

    def test_simple(self):
        tr = self.make_transport()

        @asyncio.coroutine
        def go():
            status, data = yield from tr.perform_request(
                'GET', '/_nodes/_all/clear')
            self.assertEqual(200, status)
            self.assertIn('nodes', data)
            # self.assertEqual(
            #     {'nodes':
            #      {'kagIbHGHS3a0dcyPmp0Jkw':
            #       {'version': '1.3.1',
            #        'ip': '127.0.1.1',
            #        'build': '2de6dc5',
            #        'name': 'Mandrill',
            #        'transport_address':
            #        'inet[/192.168.0.183:9300]',
            #        'http_address': 'inet[/192.168.0.183:9200]',
            #        'host': 'andrew-levelup'}},
            #      'cluster_name': 'elasticsearch'}, data)
        self.loop.run_until_complete(go())

    def test_set_endpoints(self):
        tr = self.make_transport([])
        self.assertEqual([], tr.endpoints)
        tr.endpoints = [{'host': 'localhost'}]
        self.assertEqual([Endpoint('http', 'localhost', 9200)], tr.endpoints)
        self.assertEqual(1, len(tr._pool.connections))

    def test_set_endpoints_Endpoint(self):
        tr = self.make_transport([])
        self.assertEqual([], tr.endpoints)
        tr.endpoints = [Endpoint('http', 'localhost', 9200)]
        self.assertEqual([Endpoint('http', 'localhost', 9200)], tr.endpoints)
        self.assertEqual(1, len(tr._pool.connections))

    def test_dont_recreate_existing_connections(self):
        tr = self.make_transport()
        connections = tr._pool.connections
        tr.endpoints = [{'host': 'localhost'}]
        self.assertEqual([Endpoint('http', 'localhost', 9200)], tr.endpoints)
        self.assertEqual(connections, tr._pool.connections)

    def test_set_malformed_endpoints(self):
        tr = self.make_transport()
        with self.assertRaises(RuntimeError):
            tr.endpoints = [123]
        self.assertEqual([Endpoint('http', 'localhost', 9200)], tr.endpoints)
        self.assertEqual(1, len(tr._pool.connections))

    def test_set_host_only_string(self):
        tr = self.make_transport()
        tr.endpoints = ['host']
        self.assertEqual([Endpoint('http', 'host', 9200)], tr.endpoints)
        self.assertEqual(1, len(tr._pool.connections))

    def test_set_host_port_string(self):
        tr = self.make_transport()
        tr.endpoints = ['host:123']
        self.assertEqual([Endpoint('http', 'host', 123)], tr.endpoints)
        self.assertEqual(1, len(tr._pool.connections))

    def test_set_host_port_string_invalid(self):
        tr = self.make_transport()
        with self.assertRaises(RuntimeError):
            tr.endpoints = ['host:123:abc']
        self.assertEqual([Endpoint('http', 'localhost', 9200)], tr.endpoints)
        self.assertEqual(1, len(tr._pool.connections))

    def test_set_host_dict_invalid(self):
        tr = self.make_transport()
        with self.assertRaises(RuntimeError):
            tr.endpoints = [{'a': 'b'}]
        self.assertEqual([Endpoint('http', 'localhost', 9200)], tr.endpoints)
        self.assertEqual(1, len(tr._pool.connections))

    def test_username_password_endpoints_with_port(self):
        tr = self.make_transport(endpoints=['john:doe@localhost:9200'])
        self.assertEqual([Endpoint('http', 'john:doe@localhost', 9200)],
                         tr.endpoints)

    def test_username_password_endpoints_without_port(self):
        tr = self.make_transport(endpoints=['john:doe@localhost'])
        self.assertEqual([Endpoint('http', 'john:doe@localhost', 9200)],
                         tr.endpoints)

    def test_username_password_endpoints_with_port_https(self):
        tr = self.make_transport(endpoints=['https://john:doe@localhost:9200'])
        self.assertEqual([Endpoint('https', 'john:doe@localhost', 9200)],
                         tr.endpoints)
        self.assertEqual(
            ('https', 'john:doe@localhost:9200', '/', '', '', ''),
            tuple(urllib.parse.urlparse(tr._pool.connections[0]._base_url))
        )

    def test_bad_schema(self):
        with self.assertRaises(RuntimeError):
            self.make_transport(endpoints=['s3://john:doe@localhost:9200'])

    def test_default_port_https(self):
        tr = self.make_transport(endpoints=['https://localhost'])
        self.assertEqual([Endpoint('https', 'localhost', 443)],
                         tr.endpoints)

    def test_default_port_http(self):
        tr = self.make_transport(endpoints=['http://localhost'])
        self.assertEqual([Endpoint('http', 'localhost', 9200)],
                         tr.endpoints)

    def test_sniff(self):
        tr = self.make_transport(sniffer_interval=0.001)

        @asyncio.coroutine
        def go():
            t0 = time.monotonic()
            yield from asyncio.sleep(0.001, loop=self.loop)
            yield from tr.get_connection()
            self.assertGreater(tr.last_sniff, t0)

        self.loop.run_until_complete(go())

    def test_get_connection_without_sniffing(self):
        tr = self.make_transport(sniffer_interval=1000)

        @asyncio.coroutine
        def go():
            t0 = tr.last_sniff
            yield from tr.get_connection()
            self.assertEqual(t0, tr.last_sniff)

        self.loop.run_until_complete(go())

    @unittest.expectedFailure
    def test__mark_dead(self):
        tr = self.make_transport()

        @asyncio.coroutine
        def go():
            conn = yield from tr.get_connection()
            last_sniff = tr.last_sniff
            yield from tr._mark_dead(conn)
            self.assertGreater(tr.last_sniff, last_sniff)
            self.assertEqual(1, tr._pool._dead_count[conn])

        self.loop.run_until_complete(go())

    def test_perform_request_body_bytes(self):
        tr = self.make_transport()

        @asyncio.coroutine
        def go():
            status, data = yield from tr.perform_request(
                'GET', '/_nodes/_all', body=b'')

        self.loop.run_until_complete(go())



import asyncio

from .utils import NamespacedClient
from .utils import _make_path

default = object()


def _decode_text(s):
    return s


class CatClient(NamespacedClient):

    @asyncio.coroutine
    def aliases(self, *, name=default, h=default, help=default,
                local=default, master_timeout=default, v=default):
        """
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cat-alias.html>`_

        :arg name: A comma-separated list of alias names to return
        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for connection to
            master node
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}
        if name is not default:
            params['name'] = name
        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cat', 'aliases', name),
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def allocation(self, node_id=None, *,
                   h=default, help=default, local=default,
                   master_timeout=default, v=default):
        """
        Allocation provides a snapshot of how shards have located around the
        cluster and the state of disk usage.

        """
        params = {}
        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)
        _, data = yield from self.transport.perform_request(
            'GET',
            _make_path('_cat', 'allocation', node_id),
            params=params, decoder=_decode_text)
        return data

    @asyncio.coroutine
    def count(self, index=None, *, h=default, help=default, local=default,
              master_timeout=default, v=default):
        """
        Count provides quick access to the document count of the entire
        cluster, or individual indices.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-count.html>`_

        :arg index: A comma-separated list of index names to limit the returned
            information
        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for connection to
            master node
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cat', 'count', index),
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def health(self, *, h=default, help=default, local=default,
               master_timeout=default, ts=default, v=default):
        """
        health is a terse, one-line representation of the same information from
        :meth:`~elasticsearch.client.cluster.ClusterClient.health` API
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-health.html>`_

        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for connection to
            master node
        :arg ts: Set to false to disable timestamping, default True
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if ts is not default:
            params['ts'] = bool(ts)
        if v is not default:
            params['v'] = bool(v)
        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cat', 'health'),
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def help(self, *, help=default):
        """A simple help for the cat api."""
        params = {}
        if help is not default:
            params['help'] = bool(help)
        _, data = yield from self.transport.perform_request(
            'GET', '/_cat', params=params, decoder=_decode_text)
        return data

    @asyncio.coroutine
    def indices(self, index=None, *, bytes=default, h=default, help=default,
                local=default, master_timeout=default, pri=default, v=default):
        """
        The indices command provides a cross-section of each index.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-indices.html>`_

        :arg index: A comma-separated list of index names to limit the returned
            information
        :arg bytes: The unit in which to display byte values
        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for connection to
            master node
        :arg pri: Set to true to return stats only for primary shards, default
            False
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if bytes is not default:
            params['bytes'] = bytes
        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if pri is not default:
            params['pri'] = bool(pri)
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cat', 'indices', index),
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def master(self, *, h=default, help=default, local=default,
               master_timeout=default, v=default):
        """
        Displays the master's node ID, bound IP address, and node name.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-master.html>`_

        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for connection to
            master node
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)
        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cat', 'master'),
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def nodes(self, *, h=default, help=default, local=default,
              master_timeout=default, v=default):
        """
        The nodes command shows the cluster topology.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-nodes.html>`_

        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for master connection
            node
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', '/_cat/nodes',
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def recovery(self, index=None, *, bytes=default, h=default, help=default,
                 local=default, master_timeout=default, v=default):
        """
        recovery is a view of shard replication.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-recovery.html>`_

        :arg index: A comma-separated list of index names to limit the returned
            information
        :arg bytes: The unit in which to display byte values
        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for master connection
            node
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if bytes is not default:
            params['bytes'] = bytes
        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cat', 'recovery', index),
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def shards(self, index=None, *, h=default, help=default, local=default,
               master_timeout=default, v=default):
        """
        The shards command is the detailed view of what nodes
        contain which shards.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-shards.html>`_

        :arg index: A comma-separated list of index names to limit the returned
            information
        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for master connection
            node
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cat', 'shards', index),
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def segments(self, index=None, *, h=default, help=default, local=default,
                 master_timeout=default, v=default):
        """
        The segments command is the detailed view of Lucene segments per index.

        :arg index: A comma-separated list of index names to limit the returned
            information
        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for master connection
            node
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cat', 'segments', index),
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def pending_tasks(self, *, h=default, help=default, local=default,
                      master_timeout=default, v=default):
        """
        pending_tasks provides the same information as the
        :meth:`~elasticsearch.client.cluster.ClusterClient.pending_tasks` API
        in a convenient tabular format.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-pending-tasks.html>`_

        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for master connection
            node
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', '/_cat/pending_tasks',
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def thread_pool(self, *, full_id=default, h=default, help=default,
                    local=default, master_timeout=default, v=default):
        """
        Get information about thread pools.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-thread-pool.html>`_

        :arg full_id: Enables displaying the complete node ids (default:false)
        :arg h: Comma-separated list of column names to display
        :arg help: Return help information (default: 'false')
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for master connection
            node
        :arg v: Verbose mode. Display column headers (default: 'false')

        """
        params = {}

        if full_id is not default:
            params['full_id'] = bool(full_id)
        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', '/_cat/thread_pool',
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def fielddata(self, *, fields=default, bytes=default, h=default,
                  help=default, local=default, master_timeout=default,
                  v=default):
        """
        Shows information about currently loaded fielddata on a per-node basis.
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/cat-fielddata.html>`_

        :arg fields: A comma-separated list of fields to return the fielddata
            size
        :arg bytes: The unit in which to display byte values
        :arg h: Comma-separated list of column names to display
        :arg help: Return help information (default: 'false')
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for master connection
            node
        :arg v: Verbose mode. Display column headers (default: 'false')

        """
        params = {}

        if fields is not default:
            params['fields'] = fields
        if bytes is not default:
            params['bytes'] = bytes
        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', _make_path('_cat', 'fielddata'),
            params=params, decoder=_decode_text
        )
        return data

    @asyncio.coroutine
    def plugins(self, *, h=default, help=default, local=default,
                master_timeout=default, v=default):
        """
        `<http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/cat-plugins.html>`_

        :arg h: Comma-separated list of column names to display
        :arg help: Return help information, default False
        :arg local: Return local information, do not retrieve the state from
            master node (default: false)
        :arg master_timeout: Explicit operation timeout for master connection
            node
        :arg v: Verbose mode. Display column headers, default False
        """
        params = {}

        if h is not default:
            params['h'] = h
        if help is not default:
            params['help'] = bool(help)
        if local is not default:
            params['local'] = bool(local)
        if master_timeout is not default:
            params['master_timeout'] = master_timeout
        if v is not default:
            params['v'] = bool(v)

        _, data = yield from self.transport.perform_request(
            'GET', '/_cat/plugins',
            params=params, decoder=_decode_text
        )
        return data



import asyncio
import unittest
from aioes import Elasticsearch
from aioes.exception import (NotFoundError, ConflictError,
                             RequestError, TransportError)

import pprint
pp = pprint.pprint


MESSAGES = [
    {
        "user": "Johny Mnemonic",
        "birthDate": "2109-11-15T14:12:12",
        "message": "trying out Elasticsearch",
        "skills": ["Python", "PHP", "HTML", "C++", ".NET", "JavaScript"],
        "counter": 0
    },
    {
        "user": "Sidor Spiridonovich",
        "birthDate": "2009-01-11T11:02:11",
        "message": "trying in Elasticsearch",
        "skills": ["Java", "1C", "C++", ".NET", "JavaScript"],
        "counter": 0
    },
    {
        "user": "Fedor Poligrafovich",
        "birthDate": "1969-12-15T14:12:12",
        "message": "trying out everything",
        "skills": ["MODULA", "ADA", "PLM", "BASIC", "Python"],
        "counter": 0
    },
    {
        "user": "Super",
        "birthDate": "1912-11-15T14:12:12",
        "message": "trying out ssdff  everything",
        "skills": ["MODULA", "ADA", "PLM", "BASIC", "Python"],
        "counter": 10
    },
]


class TestClient(unittest.TestCase):
    def setUp(self):
        self._index = 'test_elasticsearch'
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.cl = Elasticsearch([{'host': 'localhost'}], loop=self.loop)
        self.addCleanup(self.cl.close)
        try:
            self.loop.run_until_complete(self.cl.delete(self._index, '', ''))
        except NotFoundError:
            pass

    def tearDown(self):
        self.loop.close()

    def test_ping(self):
        """ ping """

        class R:
            @asyncio.coroutine
            def perform_request(self, a, b):
                yield
                raise TransportError

            def close(self):
                pass

        @asyncio.coroutine
        def go():
            data = yield from self.cl.ping()
            self.assertTrue(data)
            self.cl._transport = R()
            yield from self.cl.ping()
        self.loop.run_until_complete(go())
        self.cl.__repr__()

    def test_info(self):
        """ test_info """
        @asyncio.coroutine
        def go():
            data = yield from self.cl.info()
            self.assertEqual(data['status'], 200)
        self.loop.run_until_complete(go())

    def test_create(self):
        """ create index """
        @asyncio.coroutine
        def go():
            data = yield from self.cl.create(
                self._index, 'tweet',
                {
                    'user': 'Bob',
                    'skills': ['C', 'Python', 'Assembler'],
                    'date': '2009-11-15T14:12:12'
                },
                '1',
                routing='Bob')
            self.assertEqual(data['_index'], self._index)
            self.assertEqual(data['_type'], 'tweet')
            self.assertEqual(data['_version'], 1)
            self.assertTrue(data['created'], data)
            # test for conflict
            with self.assertRaises(ConflictError):
                yield from self.cl.create(
                    self._index, 'tweet', {}, '1')
        self.loop.run_until_complete(go())

    def test_index(self):
        """ auto-create index """
        @asyncio.coroutine
        def go():
            data = yield from self.cl.index(self._index, 'tweet', {}, '1')
            self.assertEqual(data['_index'], self._index)
            self.assertEqual(data['_type'], 'tweet')
            self.assertEqual(data['_id'], '1')
            self.assertEqual(data['_version'], 1)
            self.assertTrue(data['created'], data)
            # test increment version
            data = yield from self.cl.index(self._index, 'tweet', {}, '1')
            self.assertEqual(data['_version'], 2)
            self.assertFalse(data['created'], data)
            # test 'external' version_type
            data = yield from self.cl.index(self._index, 'tweet', {}, '12',
                                            version_type='external',
                                            version=122,
                                            timestamp='2009-11-15T14:12:12',
                                            ttl='1d',
                                            consistency='one',
                                            timeout=30000,
                                            refresh=True,
                                            replication='async')
            self.assertEqual(data['_version'], 122)
            self.assertTrue(data['created'], data)
            with self.assertRaises(RequestError):
                yield from self.cl.index(self._index, 'type', {},
                                         parent='1',
                                         percolate='')
            with self.assertRaises(TypeError):
                yield from self.cl.index(self._index, 'type', {},
                                         consistency=1)
            with self.assertRaises(ValueError):
                yield from self.cl.index(self._index, 'type', {},
                                         consistency='1')
            with self.assertRaises(TypeError):
                yield from self.cl.index(self._index, 'type', {},
                                         replication=1)
            with self.assertRaises(ValueError):
                yield from self.cl.index(self._index, 'type', {},
                                         replication='1')
            with self.assertRaises(TypeError):
                yield from self.cl.index(self._index, 'type', {},
                                         op_type=1)
            with self.assertRaises(ValueError):
                yield from self.cl.index(self._index, 'type', {},
                                         op_type='1')
            with self.assertRaises(TypeError):
                yield from self.cl.index(self._index, 'tweet', {},
                                         version_type=1)
            with self.assertRaises(ValueError):
                yield from self.cl.index(self._index, 'tweet', {},
                                         version_type='1')

        self.loop.run_until_complete(go())

    def test_exist(self):
        """ exists """
        @asyncio.coroutine
        def go():
            id = '100'
            # test non-exist
            data = yield from self.cl.exists(self._index, id,
                                             refresh=True,
                                             realtime=True,
                                             preference='_local')
            self.assertFalse(data)
            # test exist
            yield from self.cl.index(self._index, 'exist',
                                     {'user': 'opa', 'tim': 'none'},
                                     id,
                                     routing='opa')
            data = yield from self.cl.exists(self._index, id,
                                             routing='opa')
            self.assertTrue(data)
            data = yield from self.cl.exists(self._index, id, parent='1')
            self.assertFalse(data, data)
        self.loop.run_until_complete(go())

    def test_get(self):
        """ get """
        @asyncio.coroutine
        def go():
            id = '200'
            yield from self.cl.index(self._index, 'test_get', MESSAGES[1], id)
            data = yield from self.cl.get(self._index, id,
                                          realtime=True,
                                          refresh=True)
            self.assertEqual(data['_id'], id)
            self.assertEqual(data['_index'], self._index)
            self.assertEqual(data['_type'], 'test_get')
            self.assertEqual(data['_version'], 1)
            self.assertTrue(data['found'], data)
            self.assertEqual(data['_source'], MESSAGES[1])
            data = yield from self.cl.get(self._index, id,
                                          _source=False)
            self.assertEqual(data['_source'], {})
            data = yield from self.cl.get(self._index, id,
                                          _source_exclude='counter',
                                          _source_include='*')
            self.assertNotIn('counter', data, data)
            data = yield from self.cl.get(self._index, id,
                                          fields='user,skills',
                                          routing='Sidor',
                                          preference='_local',
                                          version=1,
                                          version_type='internal')
            self.assertIn('fields', data, data)
            self.assertIn('user', data['fields'], data)
            self.assertIn('skills', data['fields'], data)
            with self.assertRaises(NotFoundError):
                yield from self.cl.get(self._index, id, parent='1')
            with self.assertRaises(TypeError):
                yield from self.cl.get(self._index, id,
                                       version_type=1)
            with self.assertRaises(ValueError):
                yield from self.cl.get(self._index, id,
                                       version_type='1')

        self.loop.run_until_complete(go())

    def test_get_source(self):
        """ get_source """
        @asyncio.coroutine
        def go():
            yield from self.cl.index(self._index,
                                     'test_get_source',
                                     MESSAGES[0],
                                     '1')
            data = yield from self.cl.get_source(self._index, '1')
            self.assertEqual(data, MESSAGES[0])

            id = '200'
            yield from self.cl.index(
                self._index, 'test_get_source', MESSAGES[2], id,
                routing='Poligrafovich'
                )
            data = yield from self.cl.get_source(self._index, id,
                                                 routing='Poligrafovich',
                                                 preference='_local',
                                                 version=1,
                                                 version_type='internal',
                                                 realtime=True,
                                                 refresh=True)
            self.assertEqual(data, MESSAGES[2])
            data = yield from self.cl.get_source(self._index, id,
                                                 routing='Poligrafovich',
                                                 _source=False)
            self.assertEqual(data, {})
            data = yield from self.cl.get_source(self._index, id,
                                                 routing='Poligrafovich',
                                                 _source_exclude='counter',
                                                 _source_include='*')
            self.assertNotIn('counter', data, data)
            self.assertTrue('user', data)
            with self.assertRaises(NotFoundError):
                yield from self.cl.get_source(self._index, id, parent='1')
            with self.assertRaises(TypeError):
                yield from self.cl.get_source(self._index, id,
                                              version_type=1)
            with self.assertRaises(ValueError):
                yield from self.cl.get_source(self._index, id,
                                              version_type='1')
        self.loop.run_until_complete(go())

    def test_delete(self):
        """ delete """
        @asyncio.coroutine
        def go():
            yield from self.cl.index(self._index, 'testdoc', MESSAGES[2], '1')
            data = yield from self.cl.delete(self._index, 'testdoc', '1')
            self.assertTrue(data['found'], data)
            with self.assertRaises(NotFoundError):
                data = yield from self.cl.delete(self._index, 'testdoc', '1',
                                                 consistency='one',
                                                 replication='async',
                                                 refresh=True,
                                                 version_type='internal',
                                                 version=2,
                                                 timeout=30000,
                                                 routing='test',
                                                 parent='1')
            with self.assertRaises(TypeError):
                yield from self.cl.delete(self._index, 'type', {},
                                          consistency=1)
            with self.assertRaises(ValueError):
                yield from self.cl.delete(self._index, 'type', {},
                                          consistency='1')
            with self.assertRaises(TypeError):
                yield from self.cl.delete(self._index, 'type', {},
                                          replication=1)
            with self.assertRaises(ValueError):
                yield from self.cl.delete(self._index, 'type', {},
                                          replication='1')
            with self.assertRaises(TypeError):
                yield from self.cl.delete(self._index, 'type', {},
                                          version_type=1)
            with self.assertRaises(ValueError):
                yield from self.cl.delete(self._index, 'type', {},
                                          version_type='1')

        self.loop.run_until_complete(go())

    def test_update(self):
        """ update """
        @asyncio.coroutine
        def go():
            script = {
                "doc": {
                    "counter": 123
                }
            }
            yield from self.cl.index(self._index, 'testdoc', MESSAGES[2],
                                     '1',
                                     routing='Fedor')
            yield from self.cl.update(self._index, 'testdoc', '1',
                                      script,
                                      version_type='internal',
                                      version=1,
                                      routing='Fedor')
            data = yield from self.cl.get(self._index, '1', routing='Fedor')
            self.assertEqual(data['_source']['counter'], 123)
            self.assertEqual(data['_version'], 2)

            data = yield from self.cl.update(self._index, 'testdoc', '1',
                                             script,
                                             timestamp='2009-11-15T14:12:12',
                                             ttl='1d',
                                             consistency='one',
                                             timeout=30000,
                                             refresh=True,
                                             replication='async',
                                             retry_on_conflict=2,
                                             routing='Fedor',
                                             lang='en')
            self.assertEqual(data['_version'], 3)
            with self.assertRaises(NotFoundError):
                yield from self.cl.update(
                    self._index, 'testdoc', '1',
                    script={},
                    fields='user',
                    parent='1')
            with self.assertRaises(TypeError):
                yield from self.cl.update(self._index, 'type', {},
                                          consistency=1)
            with self.assertRaises(ValueError):
                yield from self.cl.update(self._index, 'type', {},
                                          consistency='1')
            with self.assertRaises(TypeError):
                yield from self.cl.update(self._index, 'type', {},
                                          replication=1)
            with self.assertRaises(ValueError):
                yield from self.cl.update(self._index, 'type', {},
                                          replication='1')
            with self.assertRaises(TypeError):
                yield from self.cl.update(self._index, 'type', {},
                                          version_type=1)
            with self.assertRaises(ValueError):
                yield from self.cl.update(self._index, 'type', {},
                                          version_type='1')

        self.loop.run_until_complete(go())

    def test_search(self):
        """ search """
        @asyncio.coroutine
        def go():
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[0], '1',
                                     refresh=True)
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[1], '2',
                                     refresh=True)
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[2], '3',
                                     refresh=True)
            data = yield from self.cl.search(self._index,
                                             'testdoc',
                                             q='skills:Python',
                                             _source=False,
                                             _source_include='skills')
            self.assertEqual(data['hits']['total'], 2, data)
            self.assertIn('skills', data['hits']['hits'][0]['_source'])
            self.assertIn('skills', data['hits']['hits'][1]['_source'])
            data = yield from self.cl.search(self._index,
                                             'testdoc',
                                             q='skills:Python',
                                             _source_exclude='skills',
                                             analyzer='standard',
                                             default_operator='AND',
                                             analyze_wildcard=True,
                                             version=2,
                                             timeout=30000,
                                             allow_no_indices=True,
                                             ignore_unavailable=True,
                                             df='_all',
                                             explain=True,
                                             fields='skills,user',
                                             from_=0,
                                             expand_wildcards='open',
                                             lenient=True,
                                             lowercase_expanded_terms=True,
                                             preference='random',
                                             scroll='1s',
                                             search_type='query_then_fetch',
                                             size=100,
                                             sort='user:true',
                                             stats=True
                                             )
            self.assertNotIn('skills', data['hits']['hits'][0]['_source'])
            self.assertNotIn('skills', data['hits']['hits'][1]['_source'])
            with self.assertRaises(TypeError):
                yield from self.cl.search(default_operator=1,
                                          indices_boost=False)
            with self.assertRaises(ValueError):
                yield from self.cl.search(doc_type='testdoc',
                                          q='skills:Python',
                                          routing='Sidor',
                                          source='Query DSL',
                                          suggest_field='user',
                                          suggest_text='test',
                                          suggest_mode='missing',
                                          suggest_size=100,
                                          default_operator='1')

            with self.assertRaises(TypeError):
                yield from self.cl.search(self._index,
                                          'testdoc',
                                          q='skills:Python',
                                          suggest_mode=1)
            with self.assertRaises(ValueError):
                yield from self.cl.search(self._index,
                                          'testdoc',
                                          q='skills:Python',
                                          suggest_mode='1')

            with self.assertRaises(TypeError):
                yield from self.cl.search(self._index,
                                          'testdoc',
                                          q='skills:Python',
                                          search_type=1)
            with self.assertRaises(ValueError):
                yield from self.cl.search(self._index,
                                          'testdoc',
                                          q='skills:Python',
                                          search_type='1')

            with self.assertRaises(TypeError):
                yield from self.cl.search(self._index,
                                          'testdoc',
                                          q='skills:Python',
                                          expand_wildcards=1)
            with self.assertRaises(ValueError):
                yield from self.cl.search(self._index,
                                          'testdoc',
                                          q='skills:Python',
                                          expand_wildcards='1')
        self.loop.run_until_complete(go())

    def test_count(self):
        """ count """
        @asyncio.coroutine
        def go():
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[0], '1',
                                     refresh=True)
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[1], '2',
                                     refresh=True)
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[2], '3',
                                     refresh=True)
            data = yield from self.cl.count(
                self._index, 'testdoc', q='skills:Python')
            self.assertEqual(data['count'], 2, data)
            data = yield from self.cl.count(
                self._index, 'testdoc', q='skills:Python',
                ignore_unavailable=True,
                expand_wildcards='open',
                allow_no_indices=False,
                min_score=1,
                preference='random')
            self.assertEqual(data['count'], 2, data)

            with self.assertRaises(TypeError):
                yield from self.cl.count(
                    self._index, 'testdoc',
                    expand_wildcards=1)

            with self.assertRaises(ValueError):
                yield from self.cl.count(
                    self._index, 'testdoc', q='skills:Python',
                    expand_wildcards='1',
                    routing='Sidor',
                    source='Query DSL')

        self.loop.run_until_complete(go())

    def test_explain(self):
        """ explain """
        @asyncio.coroutine
        def go():
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[0], '1',
                                     refresh=True)
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[1], '2',
                                     refresh=True)
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[2], '3',
                                     refresh=True)

            data = yield from self.cl.explain(
                self._index, 'testdoc', '3',
                q='skills:Python')
            self.assertTrue(data['matched'], data)
            data = yield from self.cl.explain(
                self._index, 'testdoc', '1',
                q='skills:Python',
                analyze_wildcard=True,
                _source=False,
                _source_include='user',
                _source_exclude='counter',
                analyzer='standard',
                default_operator='and',
                df='_all',
                fields='user,counter',
                lenient=True,
                lowercase_expanded_terms=False,
                preference='random')
            self.assertTrue(data['matched'], data)

            with self.assertRaises(TypeError):
                yield from self.cl.explain(
                    self._index, 'testdoc', '1',
                    q='skills:Python',
                    default_operator=1)
            with self.assertRaises(ValueError):
                yield from self.cl.explain(
                    self._index, 'testdoc', '1',
                    default_operator='1',
                    parent='2',
                    routing='Sidor',
                    source='DSL Query')

        self.loop.run_until_complete(go())

    def test_delete_by_query(self):
        """ delete_by_query """
        DQ = {"query": {"term": {"user": "Fedor Poligrafovich"}}}

        @asyncio.coroutine
        def go():
            yield from self.cl.index(self._index, 'testdoc', MESSAGES[3], '1')
            yield from self.cl.index(self._index, 'testdoc', MESSAGES[2], '2')
            # data = yield from self.cl.delete(self._index, 'testdoc', '1')
            # self.assertTrue(data['found'], data)

            data = yield from self.cl.delete_by_query(
                self._index,
                'testdoc',
                q='user:Fedor Poligrafovich'
            )
            self.assertIn('_indices', data, data)
            with self.assertRaises(TransportError):
                yield from self.cl.delete_by_query(
                    body=DQ,
                    allow_no_indices=True,
                    analyzer='standard',
                    df='_all',
                    expand_wildcards='open',
                    consistency='all',
                    default_operator='AND',
                    ignore_unavailable=True,
                    replication='async',
                    routing='Fedor',
                    source='',
                    timeout='100ms')
            with self.assertRaises(TypeError):
                yield from self.cl.delete_by_query(default_operator=1)
            with self.assertRaises(ValueError):
                yield from self.cl.delete_by_query(default_operator='1')
            with self.assertRaises(TypeError):
                yield from self.cl.delete_by_query(consistency=1)
            with self.assertRaises(ValueError):
                yield from self.cl.delete_by_query(consistency='1')
            with self.assertRaises(TypeError):
                yield from self.cl.delete_by_query(replication=1)
            with self.assertRaises(ValueError):
                yield from self.cl.delete_by_query(replication='1')
            with self.assertRaises(TypeError):
                yield from self.cl.delete_by_query(expand_wildcards=1)
            with self.assertRaises(ValueError):
                yield from self.cl.delete_by_query(expand_wildcards='1')
        self.loop.run_until_complete(go())

    def test_msearch(self):
        """ msearch """
        queries = [
            {"_index": self._index},
            {"query": {"match_all": {}}, "from": 0, "size": 10},
            {"_index": self._index},
            {"query": {"match_all": {}}}
        ]

        @asyncio.coroutine
        def go():
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[0], '1',
                                     refresh=True)
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[1], '2',
                                     refresh=True)
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[2], '3',
                                     refresh=True)

            data = yield from self.cl.msearch(queries)
            self.assertGreater(len(data['responses']), 0, data)
            data = yield from self.cl.msearch(queries, search_type='count')
            self.assertGreater(len(data['responses']), 0, data)
            with self.assertRaises(TypeError):
                yield from self.cl.msearch(queries, search_type=1)
            with self.assertRaises(ValueError):
                yield from self.cl.msearch(queries, search_type='1')

        self.loop.run_until_complete(go())

    def test_scroll(self):
        """ scroll """
        @asyncio.coroutine
        def go():
            pass
            # scroll_id = 'c2Nhbjs2OzMFXNlNyNm5JWUc1'
            # with self.assertRaises(RequestError):
            #     yield from self.cl.scroll(scroll_id,
            #                               scroll='1m')
            # with self.assertRaises(RequestError):
            #     yield from self.cl.scroll(scroll_id)

        self.loop.run_until_complete(go())

    def test_clear_scroll(self):
        """ clear_scroll """
        @asyncio.coroutine
        def go():
            pass
            # scroll_id = 'c2Nhbjs2OzMFXNlNyNm5JWUc1'
            # yield from self.cl.scroll(scroll_id)
            # yield from self.cl.clear_scroll(scroll_id)

        self.loop.run_until_complete(go())

    def test_bulk(self):
        bulks = [
            {"index": {"_index": self._index, "_type": "type1", "_id": "1"}},
            {"name": "hiq", "age": 10},
            {"index": {"_index": self._index, "_type": "type1", "_id": "2"}},
            {"name": "hiq", "age": 10},
            {"index": {"_index": self._index, "_type": "type1", "_id": "3"}},
            {"name": "hiq", "age": 10}
        ]

        @asyncio.coroutine
        def go():
            data = yield from self.cl.bulk(bulks)
            self.assertFalse(data['errors'])
            self.assertEqual(3, len(data['items']), data)
            data = yield from self.cl.bulk(
                bulks,
                consistency='one',
                refresh=True,
                routing='hiq',
                replication='async',
                timeout='1s'
            )
            with self.assertRaises(TypeError):
                yield from self.cl.bulk(bulks, consistency=1)
            with self.assertRaises(ValueError):
                yield from self.cl.bulk(bulks, consistency='1')
            with self.assertRaises(TypeError):
                yield from self.cl.bulk(bulks, replication=1)
            with self.assertRaises(ValueError):
                yield from self.cl.bulk(bulks, replication='1')
        self.loop.run_until_complete(go())

    def test_mget(self):
        """ mget """
        @asyncio.coroutine
        def go():
            yield from self.cl.index(
                self._index, 'testdoc', MESSAGES[0], '1', refresh=True)
            yield from self.cl.index(
                self._index, 'testdoc', MESSAGES[1], '2', refresh=True)
            yield from self.cl.index(
                self._index, 'testdoc', MESSAGES[2], '3', refresh=True)
            body = {
                "docs": [
                    {"_index": self._index, "_type": "testdoc", "_id": "1"},
                    {"_index": self._index, "_type": "testdoc", "_id": "2"}
                ]
            }
            data = yield from self.cl.mget(body)
            self.assertEqual(len(data['docs']), 2)
            data = yield from self.cl.mget(
                body,
                _source_exclude='birthDate',
                _source_include='user,skills',
                _source=False,
                fields='user,skills',
                realtime=True,
                refresh=True,
                preference='random',
                parent=''
            )
            self.assertIn('skills', data['docs'][0]['fields'], data)
            self.assertIn('user', data['docs'][0]['fields'], data)
            self.assertIn('skills', data['docs'][0]['_source'], data)
            self.assertIn('user', data['docs'][0]['_source'], data)
            yield from self.cl.mget(body, routing='Sidor')

        self.loop.run_until_complete(go())

    def test_suggest(self):
        """ search """
        @asyncio.coroutine
        def go():
            mapping = {
                "testdoc": {
                    "properties": {
                        "birthDate": {
                            "type": "date",
                            "format": "dateOptionalTime"
                        },
                        "counter": {
                            "type": "long"
                        },
                        # this one is different
                        "message": {
                            "type": "completion"
                        },
                        "skills": {
                            "type": "string"
                        },
                        "user": {
                            "type": "string"
                        }
                    }
                }
            }

            yield from self.cl.indices.create(self._index)
            yield from self.cl.indices.put_mapping(
                self._index,
                'testdoc',
                mapping,
            )
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[0], '1',
                                     refresh=True)
            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[1], '2',
                                     refresh=True)
            b = {
                "my-suggestion": {
                    "text": "trying out",
                    "completion": {
                        "field": "message"
                    }
                }
            }

            data = yield from self.cl.suggest(
                self._index,
                body=b,
            )
            results = data['my-suggestion'][0]['options']
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['text'], 'trying out Elasticsearch')

        self.loop.run_until_complete(go())

    def test_percolate(self):
        @asyncio.coroutine
        def go():
            mapping = {
                "testdoc": {
                    "properties": {
                        "message": {
                            "type": "string"
                        }
                    }
                }
            }
            yield from self.cl.indices.create(self._index)
            yield from self.cl.indices.put_mapping(
                self._index,
                'testdoc',
                mapping,
            )

            percolator = {
                "query": {
                    "match": {
                        "message": "bonsai tree"
                    }
                }
            }
            # register percolator
            yield from self.cl.index(self._index, '.percolator',
                                     percolator, '1',
                                     refresh=True)

            b = {
                "doc": {
                    "message": "A new bonsai tree in the office"
                }
            }
            # percolate a doc from b
            data = yield from self.cl.percolate(
                self._index,
                'testdoc',
                body=b,
            )
            self.assertEqual(data['total'], 1)
            self.assertEqual(
                data['matches'][0],
                {'_index': 'test_elasticsearch', '_id': '1'}
            )

            # percolate_count gives only count, no matches
            data = yield from self.cl.count_percolate(
                self._index,
                'testdoc',
                body=b,
            )

            self.assertEqual(data['total'], 1)
            self.assertTrue('matches' not in data)

        self.loop.run_until_complete(go())

    def test_mpercolate(self):
        @asyncio.coroutine
        def go():
            mapping = {
                "testdoc": {
                    "properties": {
                        "message": {
                            "type": "string"
                        }
                    }
                }
            }
            yield from self.cl.indices.create(self._index)
            yield from self.cl.indices.put_mapping(
                self._index,
                'testdoc',
                mapping,
            )

            percolator = {
                "query": {
                    "match": {
                        "message": "bonsai tree"
                    }
                }
            }
            # register percolator
            yield from self.cl.index(self._index, '.percolator',
                                     percolator, '1',
                                     refresh=True)

            body = [
                {
                    'percolate': {
                        'index': self._index,
                        'type': 'testdoc',
                    }
                },
                {
                    "doc": {
                        "message": "A new bonsai tree in the office"
                    }
                }
            ]

            data = yield from self.cl.mpercolate(
                body,
                self._index,
                'testdoc',
            )

            self.assertEqual(len(data['responses']), 1)
            item = data['responses'][0]
            self.assertEqual(item['total'], 1)
            self.assertEqual(
                item['matches'][0],
                {'_index': 'test_elasticsearch', '_id': '1'}
            )

        self.loop.run_until_complete(go())

    def test_termvector(self):
        @asyncio.coroutine
        def go():
            mapping = {
                "testdoc": {
                    "properties": {
                        "message": {
                            "type": "string",
                            "term_vector": "with_positions_offsets_payloads",
                            "store": True,
                        }
                    }
                }
            }
            yield from self.cl.indices.create(self._index)
            yield from self.cl.indices.put_mapping(
                self._index,
                'testdoc',
                mapping,
            )

            doc = {
                'message': 'Hello world',
            }

            yield from self.cl.index(self._index, 'testdoc',
                                     doc, '1',
                                     refresh=True)

            data = yield from self.cl.termvector(self._index, 'testdoc', '1')

            vector_data = data['term_vectors']['message']
            self.assertEqual(vector_data['field_statistics'], {
                "sum_doc_freq": 2,
                "doc_count": 1,
                "sum_ttf": 2
            })
            self.assertTrue('hello' in vector_data['terms'])
            self.assertTrue('world' in vector_data['terms'])

        self.loop.run_until_complete(go())

    def test_mtermvectors(self):
        @asyncio.coroutine
        def go():
            mapping = {
                "testdoc": {
                    "properties": {
                        "message": {
                            "type": "string",
                            "term_vector": "with_positions_offsets_payloads",
                            "store": True,
                        }
                    }
                }
            }
            yield from self.cl.indices.create(self._index)
            yield from self.cl.indices.put_mapping(
                self._index,
                'testdoc',
                mapping,
            )

            doc = {
                'message': 'Hello world',
            }

            yield from self.cl.index(self._index, 'testdoc',
                                     doc, '1',
                                     refresh=True)
            doc = {
                'message': 'Second term',
            }

            yield from self.cl.index(self._index, 'testdoc',
                                     doc, '2',
                                     refresh=True)

            data = yield from self.cl.mtermvectors(
                self._index, 'testdoc', ids='1,2'
            )

            self.assertEqual(len(data['docs']), 2)
            self.assertTrue('term_vectors' in data['docs'][0])
            self.assertTrue('term_vectors' in data['docs'][1])

        self.loop.run_until_complete(go())

    def test_scripts_management(self):
        @asyncio.coroutine
        def go():
            script = {'script': 'log(_score * 2)'}

            # adding
            yield from self.cl.put_script('groovy', 'test_script', script)

            # getting and checking
            got_script = yield from self.cl.get_script('groovy', 'test_script')
            self.assertEqual(script['script'], got_script['script'])

            # deleting
            yield from self.cl.delete_script('groovy', 'test_script')
            with self.assertRaises(NotFoundError):
                got_script = yield from self.cl.get_script(
                    'groovy', 'test_script'
                )

        self.loop.run_until_complete(go())

    def test_scripts_execution(self):
        @asyncio.coroutine
        def go():
            script = {
                'script': '2*val',
            }
            query = {
                "query": {
                    "match": {
                        "user": "Johny Mnemonic"
                    }
                },
                "script_fields": {
                    "test1": {
                        "lang": "groovy",
                        "script_id": "calculate-score",
                        "params": {
                            "val": 2,
                        }
                    }
                }
            }

            yield from self.cl.index(self._index, 'testdoc',
                                     MESSAGES[0], '1',
                                     refresh=True)

            yield from self.cl.put_script('groovy', 'calculate-score', script)
            data = yield from self.cl.search(self._index, 'testdoc', query)
            res = data['hits']['hits'][0]['fields']['test1'][0]
            self.assertEqual(res, 4)  # 2*2

        self.loop.run_until_complete(go())

    def test_templates_management(self):
        @asyncio.coroutine
        def go():
            template = {
                "template": {
                    "query": {
                        "match": {
                            "user": "{{query_string}}"
                        }
                    }
                }
            }

            yield from self.cl.put_template('test_template', template)

            data = yield from self.cl.get_template('test_template')
            self.assertTrue(template, data)

            yield from self.cl.delete_template('test_template')
            with self.assertRaises(NotFoundError):
                yield from self.cl.get_template('test_template')

        self.loop.run_until_complete(go())

    def test_template_search(self):
        @asyncio.coroutine
        def go():
            template = {
                "template": {
                    "query": {
                        "match": {
                            "user": "{{query_string}}"
                        }
                    }
                }
            }
            search_body = {
                "template": {
                    "id": "test_template"
                },
                "params": {
                    "query_string": "Johny Mnemonic"
                }
            }
            yield from self.cl.index(
                self._index, 'testdoc', MESSAGES[0], '1',
                refresh=True
            )

            yield from self.cl.put_template('test_template', template)

            data = yield from self.cl.search_template(
                self._index, 'testdoc', body=search_body
            )
            self.assertEqual(data['hits']['total'], 1)

        self.loop.run_until_complete(go())

    def test_search_shards(self):
        @asyncio.coroutine
        def go():
            yield from self.cl.index(
                self._index, 'testdoc', MESSAGES[0], '1',
                refresh=True
            )
            data = yield from self.cl.search_shards(
                self._index, 'testdoc'
            )
            self.assertTrue('nodes' in data)
            self.assertTrue(len(data['nodes']) > 0)
            self.assertTrue('shards' in data)
            self.assertTrue(len(data['shards']) > 0)

        self.loop.run_until_complete(go())

    def test_mlt(self):
        @asyncio.coroutine
        def go():
            msg = MESSAGES[0].copy()
            # mlt needs quite a lot of text to work
            msg['message'] = '''
            Additionally, More Like This can find documents that are "like"
            a set of chosen documents. The syntax to specify one or more
            documents is similar to the Multi GET API, and supports the
            ids or docs array. If only one document is specified,
            the query behaves the same as the More Like This API.'''
            # and quite a lot of documents
            for i in range(50):
                yield from self.cl.index(
                    self._index, 'testdoc', msg, str(i),
                    refresh=True
                )

            data = yield from self.cl.mlt(
                self._index, 'testdoc', '1'
            )
            # initial document is not included
            self.assertEqual(data['hits']['total'], 49)
            self.assertEqual(data['hits']['hits'][0]['_source'], msg)

        self.loop.run_until_complete(go())

    # def test_(self):
    #     @asyncio.coroutine
    #     def go():
    #         pass
    #     self.loop.run_until_complete(go())
